# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO
from typing import Union, Tuple, Iterable

import numpy as np
import pandas as pd
from skimage.draw import polygon
from skimage.transform import resize

from . import cortical_profile
from .. import anchor as _anchor
from ...commons import logger
from ...retrieval import requests


def cell_reader(bytes_buffer: bytes):
    return pd.read_csv(BytesIO(bytes_buffer[2:]), delimiter=" ", header=0).astype(
        {"layer": int, "label": int}
    )


def layer_reader(bytes_buffer: bytes):
    return pd.read_csv(BytesIO(bytes_buffer[2:]), delimiter=" ", header=0, index_col=0)


def poly_srt(poly):
    return poly[poly[:, 0].argsort(), :]


def poly_rev(poly):
    return poly[poly[:, 0].argsort()[::-1], :]


class PolyLine:
    """Simple polyline representation which allows equidistant sampling."""

    def __init__(self, pts):
        self.pts = pts
        self.lengths = [
            np.sqrt(np.sum((pts[i, :] - pts[i - 1, :]) ** 2))
            for i in range(1, pts.shape[0])
        ]

    def length(self):
        return sum(self.lengths)

    def sample(self, d: Union[Iterable[float], np.ndarray, float]):
        # if d is iterable, we assume a list of sample positions
        try:
            iter(d)
        except TypeError:
            positions = [d]
        else:
            positions = d

        samples = []
        for s_ in positions:
            s = min(max(s_, 0), 1)
            target_distance = s * self.length()
            current_distance = 0
            for i, length in enumerate(self.lengths):
                current_distance += length
                if current_distance >= target_distance:
                    p1 = self.pts[i, :]
                    p2 = self.pts[i + 1, :]
                    r = (target_distance - current_distance + length) / length
                    samples.append(p1 + (p2 - p1) * r)
                    break

        if len(samples) == 1:
            return samples[0]
        else:
            return np.array(samples)


class CellDensityProfile(
    cortical_profile.CorticalProfile,
    configuration_folder="features/tabular/corticalprofiles/celldensity",
    category='cellular'
):

    DESCRIPTION = (
        "Cortical profile of estimated densities of detected cell bodies (in detected cells per 0.1 cube millimeter) "
        "obtained by applying a Deep Learning based instance segmentation algorithm (Contour Proposal Network; Upschulte "
        "et al., Neuroimage 2022) to a 1 micron resolution cortical image patch prepared with modified Silver staining. "
        "Densities have been computed per cortical layer after manual layer segmentation, by dividing the number of "
        "detected cells in that layer with the area covered by the layer. Therefore, each profile contains 6 measurement points. "
        "The cortical depth is estimated from the measured layer thicknesses."
    )

    BIGBRAIN_VOLUMETRIC_SHRINKAGE_FACTOR = 1.931

    _filter_attrs = cortical_profile.CorticalProfile._filter_attrs + ["location"]

    def __init__(
        self,
        section: int,
        patch: int,
        url: str,
        anchor: _anchor.AnatomicalAnchor,
        datasets: list = [],
        id: str = None,
        prerelease: bool = False,
    ):
        """
        Generate a cell density profile from a URL to a cloud folder
        formatted according to the structure used by Bludau/Dickscheid et al.
        """
        cortical_profile.CorticalProfile.__init__(
            self,
            description=self.DESCRIPTION,
            modality="Segmented cell body density",
            unit="cells / 0.1mm3",
            anchor=anchor,
            datasets=datasets,
            id=id,
            prerelease=prerelease,
        )
        self._step = 0.01
        self._url = url
        self._cell_loader = requests.HttpRequest(url, cell_reader)
        self._layer_loader = requests.HttpRequest(
            url.replace("segments", "layerinfo"), layer_reader
        )
        self._density_image = None
        self._layer_mask = None
        self._depth_image = None
        self.section = section
        self.patch = patch

    @property
    def location(self):
        return self.anchor.location

    @property
    def shape(self):
        """(y,x)"""
        return tuple(np.ceil(self.cells[["y", "x"]].max()).astype("int"))

    def boundary_annotation(self, boundary: Tuple[int, int]) -> np.ndarray:
        """Returns the annotation of a specific layer boundary."""
        shape_y, shape_x = self.shape

        # start of image patch
        if boundary == (-1, 0):
            return np.array([[0, 0], [shape_x, 0]])

        # end of image patch
        if boundary == (7, 8):
            return np.array([[0, shape_y], [shape_x, shape_y]])

        # retrieve polygon
        basename = "{}_{}.json".format(
            *(self.LAYERS[layer] for layer in boundary)
        ).replace("0_I", "0")
        poly_url = self._url.replace("segments.txt", basename)
        poly = poly_srt(np.array(requests.HttpRequest(poly_url).get()["segments"]))

        # ensure full width and trim to the image shape
        poly[0, 0] = 0
        poly[poly[:, 0] > shape_x, 0] = shape_x
        poly[poly[:, 1] > shape_y, 1] = shape_y

        return poly

    def layer_annotation(self, layer: int) -> np.ndarray:
        return np.vstack(
            (
                self.boundary_annotation((layer - 1, layer)),
                poly_rev(self.boundary_annotation((layer, layer + 1))),
                self.boundary_annotation((layer - 1, layer))[0, :],
            )
        )

    @property
    def layer_mask(self) -> np.ndarray:
        """Generates a layer mask from boundary annotations."""
        if self._layer_mask is None:
            self._layer_mask = np.zeros(np.array(self.shape, dtype=int) + 1, dtype="int")
            for layer in range(1, 8):
                pl = self.layer_annotation(layer)
                X, Y = polygon(pl[:, 0], pl[:, 1])
                self._layer_mask[Y, X] = layer
        return self._layer_mask

    @property
    def depth_image(self) -> np.ndarray:
        """Cortical depth image from layer boundary polygons by equidistant sampling."""

        if self._depth_image is None:
            logger.info("Calculating cell densities from cell and layer data...")
            # compute equidistant cortical depth image from inner and outer contour
            scale = 0.1
            depth_arr = np.zeros(np.ceil(np.array(self.shape) * scale).astype("int") + 1)

            # determine sufficient stepwidth for profile sampling
            # to match downscaled image resolution
            vstep, hstep = 1.0 / np.array(depth_arr.shape) / 2.0
            vsteps = np.arange(0, 1 + vstep, vstep)
            hsteps = np.arange(0, 1 + hstep, hstep)

            # build straight profiles between outer and inner cortical boundary
            s0 = PolyLine(self.boundary_annotation((0, 1)) * scale).sample(hsteps)
            s1 = PolyLine(self.boundary_annotation((6, 7)) * scale).sample(hsteps)
            profiles = [PolyLine(_.reshape(2, 2)) for _ in np.hstack((s0, s1))]

            # write sample depths to their location in the depth image
            for prof in profiles:
                prof_samples_as_index = prof.sample(vsteps).astype("int")
                depth_arr[prof_samples_as_index[:, 1], prof_samples_as_index[:, 0]] = vsteps

            # fix wm region, account for rounding error
            XY = self.layer_annotation(7) * scale
            depth_arr[polygon(XY[:, 1] - 1, XY[:, 0])] = 1
            depth_arr[-1, :] = 1

            # rescale depth image to original patch size
            self._depth_image = resize(depth_arr, self.density_image.shape)

        return self._depth_image

    @property
    def boundary_positions(self):
        if self._boundary_positions is None:
            self._boundary_positions = {}
            for b in self.BOUNDARIES:
                XY = self.boundary_annotation(b).astype("int")
                self._boundary_positions[b] = self.depth_image[
                    XY[:, 1], XY[:, 0]
                ].mean()
        return self._boundary_positions

    @property
    def density_image(self) -> np.ndarray:
        if self._density_image is None:
            logger.debug("Computing density image for", self._url)
            # we integrate cell counts into 2D bins
            # of square shape with a fixed sidelength
            pixel_size_micron = 100
            counts, xedges, yedges = np.histogram2d(
                self.cells.y,
                self.cells.x,
                bins=np.round(np.array(self.shape) / pixel_size_micron).astype("int"),
            )

            # rescale the counts from count / pixel_size**2  to count / 0.1mm^3,
            # assuming 20 micron section thickness.
            counts = counts / pixel_size_micron ** 2 / 20 * 100 ** 3

            # apply the Bigbrain shrinkage factor
            # TODO The planar correction factor was used for the datasets, but should
            # clarify if the full volumetric correction factor is not more correct.
            counts /= np.cbrt(self.BIGBRAIN_VOLUMETRIC_SHRINKAGE_FACTOR) ** 2

            # to go to 0.1 millimeter cube, we multiply by 0.1 / 0.0002 = 500
            self._density_image = resize(counts, self.layer_mask.shape, order=2)

        return self._density_image

    @property
    def cells(self) -> pd.DataFrame:
        return self._cell_loader.get()

    @property
    def layers(self) -> pd.DataFrame:
        return self._layer_loader.get()

    @property
    def _depths(self):
        return np.arange(0, 1, self._step) + self._step / 2

    @property
    def _values(self):
        # TODO: release a dataset update instead of on the fly computation
        densities = []
        delta = self._step / 2.0
        for d in self._depths:
            mask = (self.depth_image >= d - delta) & (self.depth_image < d + delta)
            if np.sum(mask) > 0:
                densities.append(self.density_image[mask].mean())
            else:
                densities.append(np.nan)
        return np.asanyarray(densities)



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
from textwrap import wrap

from . import tabular, cell_reader, layer_reader
from .. import anchor as _anchor
from ... import commons
from ...retrieval import requests


class LayerwiseCellDensity(
    tabular.Tabular,
    configuration_folder="features/tabular/layerstatistics/celldensity",
    category='cellular'
):

    DESCRIPTION = (
        "Layerwise estimated densities of detected cell bodies  (in detected cells per 0.1 cube millimeter) "
        "obtained by applying a Deep Learning based instance segmentation algorithm (Contour Proposal Network; Upschulte "
        "et al., Neuroimage 2022) to a 1 micron resolution cortical image patch prepared with modified Silver staining. "
        "Densities have been computed per cortical layer after manual layer segmentation, by dividing the number of "
        "detected cells in that layer with the area covered by the layer. Therefore, each profile contains 6 measurement points. "
        "The cortical depth is estimated from the measured layer thicknesses."
    )
    BIGBRAIN_VOLUMETRIC_SHRINKAGE_FACTOR = 1.931

    def __init__(
        self,
        segmentfiles: list,
        layerfiles: list,
        anchor: _anchor.AnatomicalAnchor,
        datasets: list = [],
        id: str = None,
        prerelease: bool = False,
    ):
        tabular.Tabular.__init__(
            self,
            description=self.DESCRIPTION,
            modality="Cell body density",
            anchor=anchor,
            datasets=datasets,
            data=None,  # lazy loading below
            id=id,
            prerelease=prerelease,
        )
        self.unit = "# detected cells / $0.1mm^3$"
        self._filepairs = list(zip(segmentfiles, layerfiles))
        self._densities = None

    def _load_densities(self):
        data = []
        for cellfile, layerfile in self._filepairs:
            try:
                cells = requests.HttpRequest(cellfile, func=cell_reader).data
                layers = requests.HttpRequest(layerfile, func=layer_reader).data
            except requests.SiibraHttpRequestError as e:
                print(str(e))
                commons.logger.error(f"Skipping to bootstrap a {self.__class__.__name__} feature, cannot access file resource.")
                continue
            counts = cells.layer.value_counts()
            # compute the volumetric shrinkage corrections in the same ways as it was used
            # for the pdf reports in the underlying dataset
            shrinkage_volumetric = self.BIGBRAIN_VOLUMETRIC_SHRINKAGE_FACTOR
            layer_volumes = (
                layers["Area(micron**2)"]  # this is the number of pixels, shrinkage corrected from the dataset
                * 20  # go to cube micrometer in one patch with 20 micron thickness
                * np.cbrt(shrinkage_volumetric)  # compensate linear shrinkage for 3rd dimension
                / 100 ** 3  # go to 0.1 cube millimeter
            )
            fields = cellfile.split("/")
            for layer in layer_volumes.index:
                try:
                    data.append({
                        'layer': layer,
                        'layername': layers["Name"].loc[layer],
                        'counts': counts.loc[layer],
                        'area_mu2': layers["Area(micron**2)"].loc[layer],
                        'volume': layer_volumes.loc[layer],
                        'density': counts.loc[layer] / layer_volumes.loc[layer],
                        'regionspec': fields[-5],
                        'section': int(fields[-3]),
                        'patch': int(fields[-2]),
                    })
                except KeyError:
                    continue
        return pd.DataFrame(data)

    @property
    def data(self):
        if self._data_cached is None:
            self._data_cached = self._load_densities()
            # self._data_cached.index.name = 'layer'
        return self._data_cached

    def plot(self, *args, backend="matplotlib", **kwargs):
        wrapwidth = kwargs.pop("textwrap") if "textwrap" in kwargs else 40
        kwargs["title"] = kwargs.pop(
            "title",
            "\n".join(wrap(
                f"{self.modality} in {self.anchor._regionspec or self.anchor.location}",
                wrapwidth
            ))
        )
        kwargs["kind"] = kwargs.get("kind", "box")
        kwargs["ylabel"] = kwargs.get(
            "ylabel",
            f"\n{self.unit}" if hasattr(self, 'unit') else ""
        )
        if backend == "matplotlib":
            if kwargs["kind"] == "box":
                from matplotlib.pyplot import tight_layout

                np.random.seed(int(self.data["density"].mean()))

                title = kwargs.pop("title")
                default_kwargs = {
                    "grid": True,
                    'by': "layername",
                    'column': ['density'],
                    'showfliers': False,
                    'xlabel': 'layer',
                    'color': 'dimgray',
                }
                ax, *_ = self.data.plot(*args, backend=backend, **{**default_kwargs, **kwargs})
                for i, (layer, d) in enumerate(self.data.groupby('layername')):
                    ax.scatter(
                        np.random.normal(i + 1, 0.05, len(d.density)),
                        d.density,
                        c='b', s=3
                    )
                ax.set_title(title)
                tight_layout()
                return ax
            return self.data.plot(*args, backend=backend, **kwargs)
        elif backend == "plotly":
            kwargs["title"] = kwargs["title"].replace('\n', "<br>")
            yaxis_title = kwargs.pop("ylabel")
            fig = self.data.plot(y='density', x='layer', points="all", backend=backend, **kwargs)
            fig.update_layout(yaxis_title=yaxis_title)
            return fig
        else:
            return self.data.plot(*args, backend=backend, **kwargs)



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from textwrap import wrap
from typing import List
try:
    from typing import TypedDict
except ImportError:
    from typing_extensions import TypedDict

import pandas as pd

from . import tabular
from .. import anchor as _anchor
from ...retrieval.datasets import GenericDataset


class GeneExpressions(
    tabular.Tabular,
    category='molecular'
):
    """
    A set gene expressions for different candidate genes
    measured inside a brain structure.
    """

    DESCRIPTION = """
    Gene expressions extracted from microarray data in the Allen Atlas.
    """

    ALLEN_ATLAS_NOTIFICATION = """
    For retrieving microarray data, siibra connects to the web API of
    the Allen Brain Atlas (© 2015 Allen Institute for Brain Science),
    available from https://brain-map.org/api/index.html. Any use of the
    microarray data needs to be in accordance with their terms of use,
    as specified at https://alleninstitute.org/legal/terms-use/.
    """

    DATASET = GenericDataset(
        name="An anatomically comprehensive atlas of the adult human brain transcriptome",
        contributors=[
            'Michael J. Hawrylycz',
            'Ed S. Lein',
            'Angela L. Guillozet-Bongaarts',
            'Elaine H. Shen',
            'Lydia Ng',
            'Jeremy A. Miller',
            'Louie N. van de Lagemaat',
            'Kimberly A. Smith',
            'Amanda Ebbert',
            'Zackery L. Riley',
            'Chris Abajian',
            'Christian F. Beckmann',
            'Amy Bernard',
            'Darren Bertagnolli',
            'Andrew F. Boe',
            'Preston M. Cartagena',
            'M. Mallar Chakravarty',
            'Mike Chapin',
            'Jimmy Chong',
            'Rachel A. Dalley',
            'Barry David Daly',
            'Chinh Dang',
            'Suvro Datta',
            'Nick Dee',
            'Tim A. Dolbeare',
            'Vance Faber',
            'David Feng',
            'David R. Fowler',
            'Jeff Goldy',
            'Benjamin W. Gregor',
            'Zeb Haradon',
            'David R. Haynor',
            'John G. Hohmann',
            'Steve Horvath',
            'Robert E. Howard',
            'Andreas Jeromin',
            'Jayson M. Jochim',
            'Marty Kinnunen',
            'Christopher Lau',
            'Evan T. Lazarz',
            'Changkyu Lee',
            'Tracy A. Lemon',
            'Ling Li',
            'Yang Li',
            'John A. Morris',
            'Caroline C. Overly',
            'Patrick D. Parker',
            'Sheana E. Parry',
            'Melissa Reding',
            'Joshua J. Royall',
            'Jay Schulkin',
            'Pedro Adolfo Sequeira',
            'Clifford R. Slaughterbeck',
            'Simon C. Smith',
            'Andy J. Sodt',
            'Susan M. Sunkin',
            'Beryl E. Swanson',
            'Marquis P. Vawter',
            'Derric Williams',
            'Paul Wohnoutka',
            'H. Ronald Zielke',
            'Daniel H. Geschwind',
            'Patrick R. Hof',
            'Stephen M. Smith',
            'Christof Koch',
            'Seth G. N. Grant',
            'Allan R. Jones'
        ],
        url="https://doi.org/10.1038%2Fnature11405",
        description='Neuroanatomically precise, genome-wide maps of transcript '
            'distributions are critical resources to complement genomic '
            'sequence data and to correlate functional and genetic brain '
            'architecture. Here we describe the generation and analysis '
            'of a transcriptional atlas of the adult human brain, '
            'comprising extensive histological analysis and comprehensive '
            'microarray profiling of ~900 neuroanatomically precise '
            'subdivisions in two individuals. Transcriptional regulation '
            'varies enormously by anatomical location, with different '
            'regions and their constituent cell types displaying robust '
            'molecular signatures that are highly conserved between '
            'individuals. Analysis of differential gene expression and '
            'gene co-expression relationships demonstrates that brain-'
            'wide variation strongly reflects the distributions of major '
            'cell classes such as neurons, oligodendrocytes, astrocytes '
            'and microglia. Local neighbourhood relationships between '
            'fine anatomical subdivisions are associated with discrete '
            'neuronal subtypes and genes involved with synaptic '
            'transmission. The neocortex displays a relatively '
            'homogeneous transcriptional pattern, but with distinct '
            'features associated selectively with primary sensorimotor '
            'cortices and with enriched frontal lobe expression. Notably, '
            'the spatial topography of the neocortex is strongly '
            'reflected in its molecular topography— the closer two '
            'cortical regions, the more similar their transcriptomes. '
            'This freely accessible online data resource forms a high-'
            'resolution transcriptional baseline for neurogenetic studies '
            'of normal and abnormal human brain function.'
            ""
            "For retrieving microarray data, siibra connects to the web API of "
            "the Allen Brain Atlas (© 2015 Allen Institute for Brain Science), "
            "available from https://brain-map.org/api/index.html. Any use of the "
            "microarray data needs to be in accordance with their terms of use, "
            "as specified at https://alleninstitute.org/legal/terms-use/."
    )

    class _DonorDict(TypedDict):
        id: int
        name: str
        race: str
        age: int
        gender: str

    class _SampleStructure(TypedDict):
        id: int
        name: str
        abbreviation: str
        color: str

    def __init__(
        self,
        levels: List[float],
        z_scores: List[float],
        genes: List[str],
        additional_columns: dict,
        anchor: _anchor.AnatomicalAnchor,
        datasets: List = [DATASET]
    ):
        """
        Construct gene expression table.

        Parameters
        ----------
        levels : list of float
            Expression levels measured
        z_scores : list of float
            corresponding z scores measured
        genes : list of str
            Name of the gene corresponding to each measurement
        additional_columns : dict of list
            columns with additional data to be added to the tabular feature.
            Keys give column names, values are lists with the column data.
            Each list given needs to have the same length as expression_levels
        anchor: AnatomicalAnchor
        datasets : list
            list of datasets corresponding to this feature
        """
        assert len(z_scores) == len(levels)
        assert len(genes) == len(levels)
        if additional_columns is not None:
            assert all(len(lst) == len(levels) for lst in additional_columns.values())
        else:
            additional_columns = {}

        _data_cahced = pd.DataFrame(
            dict(
                **{'level': levels, 'zscore': z_scores, 'gene': genes},
                **additional_columns
            )
        )
        # data.index.name = 'probe_id'
        tabular.Tabular.__init__(
            self,
            description=(
                (self.DESCRIPTION + self.ALLEN_ATLAS_NOTIFICATION)
                .replace('\n', ' ')
                .replace('\t', '')
                .strip()
            ),
            modality="Gene expression",
            anchor=anchor,
            data=_data_cahced,
            datasets=datasets
        )
        self.unit = "expression level"
        self._genes = list(set(genes))

    def plot(self, *args, backend="matplotlib", **kwargs):
        """
        Create a box plot per gene.

        Parameters
        ----------
        backend: str
            "matplotlib", "plotly", or others supported by pandas DataFrame
            plotting backend.
        **kwargs
            Keyword arguments are passed on to the plot command.
        """
        wrapwidth = kwargs.pop("textwrap") if "textwrap" in kwargs else 40
        kwargs["title"] = kwargs.pop(
            "title",
            "\n".join(wrap(
                f"{self.modality}\n{self.anchor._regionspec or self.anchor.location}",
                wrapwidth
            ))
        )
        kwargs["kind"] = kwargs.get("kind", "box")
        if backend == "matplotlib":
            if kwargs["kind"] == "box":
                from matplotlib.pyplot import tight_layout

                title = kwargs.pop("title")
                default_kwargs = {
                    "grid": True,
                    'by': "gene",
                    'column': ['level'],
                    'showfliers': False,
                    'ylabel': 'expression level',
                    'xlabel': 'gene',
                    'color': 'dimgray',
                    'rot': 90 if len(self._genes) > 1 else 0,
                }
                ax, *_ = self.data.plot(*args, backend=backend, **{**default_kwargs, **kwargs})
                ax.set_title(title)
                tight_layout()
                return ax
            return self.data.plot(*args, backend=backend, **kwargs)
        elif backend == "plotly":
            kwargs["title"] = kwargs["title"].replace('\n', "<br>")
            return self.data.plot(y='level', x='gene', backend=backend, **kwargs)
        else:
            return self.data.plot(*args, backend=backend, **kwargs)



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from zipfile import ZipFile
from typing import Callable, Union, List, Tuple, Iterator
try:
    from typing import Literal
except ImportError:  # support python 3.7
    from typing_extensions import Literal

import numpy as np
import pandas as pd

from .. import anchor as _anchor
from ..feature import Feature, Compoundable
from ..tabular.tabular import Tabular
from ...commons import logger, QUIET, siibra_tqdm
from ...core import region as _region
from ...locations import pointcloud
from ...retrieval.repositories import RepositoryConnector
from ...retrieval.requests import HttpRequest


class RegionalConnectivity(Feature, Compoundable):
    """
    Parcellation-averaged connectivity, providing one or more matrices of a
    given modality for a given parcellation.
    """

    _filter_attrs = ["modality", "cohort", "subject"]
    _compound_attrs = ["modality", "cohort"]

    def __init__(
        self,
        cohort: str,
        modality: str,
        regions: list,
        connector: RepositoryConnector,
        decode_func: Callable,
        filename: str,
        anchor: _anchor.AnatomicalAnchor,
        description: str = "",
        datasets: list = [],
        subject: str = "average",
        feature: str = None,
        id: str = None,
        prerelease: bool = False,
    ):
        """
        Construct a parcellation-averaged connectivity matrix.

        Parameters
        ----------
        cohort: str
            Name of the cohort used for computing the connectivity.
        modality: str
            Connectivity modality, typically set by derived classes.
        regions: list[str]
            Names of the regions from the parcellation
        connector: Connector
            Repository connector for loading the actual data array(s).
        decode_func: function
            Function to convert the bytestream of a loaded file into an array
        files: dict
            A dictionary linking names of matrices (typically subject ids)
            to the relative filenames of the data array(s) in the repository connector.
        anchor: AnatomicalAnchor
            anatomical localization of the matrix, expected to encode the parcellation
            in the region attribute.
        description: str, optional
            textual description of this connectivity matrix.
        datasets : list[Dataset]
            list of datasets corresponding to this feature.
        """
        Feature.__init__(
            self,
            modality=modality,
            description=description,
            anchor=anchor,
            datasets=datasets,
            id=id,
            prerelease=prerelease
        )
        self.cohort = cohort.upper()
        if isinstance(connector, str) and connector:
            self._connector = HttpRequest(connector, decode_func)
        else:
            self._connector = connector
        self._filename = filename
        self._decode_func = decode_func
        self.regions = regions
        self._matrix = None
        self._subject = subject
        self._feature = feature
        self._matrix_std = None  # only used for compound feature

    @property
    def subject(self):
        """Returns the subject identifiers for which the matrix represents."""
        return self._subject

    @property
    def feature(self):
        """If applicable, returns the type of feature for which the matrix represents."""
        return self._feature

    @property
    def name(self):
        return f"{self.feature or self.subject} - " + super().name + f" cohort: {self.cohort}"

    @property
    def data(self) -> pd.DataFrame:
        """
        Returns a matrix as a pandas dataframe.

        Returns
        -------
        pd.DataFrame
            A square matrix with region names as the column and row names.
        """
        if self._matrix is None:
            self._load_matrix()
        return self._matrix.copy()

    @classmethod
    def _merge_elements(
        cls,
        elements: List["RegionalConnectivity"],
        description: str,
        modality: str,
        anchor: _anchor.AnatomicalAnchor,
    ):
        assert len({f.cohort for f in elements}) == 1
        merged = cls(
            cohort=elements[0].cohort,
            regions=elements[0].regions,
            connector=elements[0]._connector,
            decode_func=elements[0]._decode_func,
            filename="",
            subject="average",
            feature="average",
            description=description,
            modality=modality,
            anchor=anchor,
            **{"paradigm": "averaged (by siibra)"} if getattr(elements[0], "paradigm", None) else {}
        )
        if isinstance(elements[0]._connector, HttpRequest):
            getter = lambda elm: elm._connector.get()
        else:
            getter = lambda elm: elm._connector.get(elm._filename, decode_func=elm._decode_func)
        all_arrays = [
            getter(elm)
            for elm in siibra_tqdm(
                elements,
                total=len(elements),
                desc=f"Averaging {len(elements)} connectivity matrices"
            )
        ]
        merged._matrix = elements[0]._arraylike_to_dataframe(
            np.stack(all_arrays).mean(0)
        )
        merged._matrix_std = elements[0]._arraylike_to_dataframe(
            np.stack(all_arrays).std(0)
        )
        return merged

    def _plot_matrix(
        self, regions: List[str] = None,
        logscale: bool = False, *args, backend="nilearn", **kwargs
    ):
        """
        Plots the heatmap of the connectivity matrix using nilearn.plotting.

        Parameters
        ----------
        regions: list[str]
            Display the matrix only for selected regions. By default, shows all the regions.
            It can only be a subset of regions of the feature.
        logscale: bool
            Display the data in log10 scale
        backend: str
            "nilearn" or "plotly"
        **kwargs:
            Can take all the arguments `nilearn.plotting.plot_matrix` can take. See the doc at
            https://nilearn.github.io/stable/modules/generated/nilearn.plotting.plot_matrix.html
        """
        if regions is None:
            regions = self.regions
        indices = [self.regions.index(r) for r in regions]
        matrix = self.data.iloc[indices, indices].to_numpy()  # nilearn.plotting.plot_matrix works better with a numpy array

        if logscale:
            matrix = np.log10(matrix)

        # default kwargs
        kwargs["title"] = kwargs.get(
            "title",
            "".join([
                f"{self.feature if self.feature else ''} - {self.subject} - ",
                f"{self.modality} in ",
                f"{', '.join({_.name for _ in self.anchor.regions})}"
            ])
        )

        if kwargs.get("reorder") or (backend == "nilearn"):
            kwargs["figure"] = kwargs.get("figure", (15, 15))
            from nilearn import plotting
            plotting.plot_matrix(
                matrix,
                labels=regions,
                **kwargs
            )
        elif backend == "plotly":
            kwargs["title"] = kwargs["title"].replace("\n", "<br>")
            from plotly.express import imshow
            return imshow(matrix, *args, x=regions, y=regions, **kwargs)
        else:
            raise NotImplementedError(
                f"Plotting connectivity matrices with {backend} is not supported."
            )

    def _to_zip(self, fh: ZipFile):
        super()._to_zip(fh)
        if self.feature is None:
            fh.writestr(f"sub/{self._filename}/matrix.csv", self.data.to_csv())
        else:
            fh.writestr(f"feature/{self._filename}/matrix.csv", self.data.to_csv())

    def get_profile(
        self,
        region: Union[str, _region.Region],
        min_connectivity: float = 0,
        max_rows: int = None,
        direction: Literal['column', 'row'] = 'column'
    ):
        """
        Extract a regional profile from the matrix, to obtain a tabular data
        feature with the connectivity as the single column. Rows are be sorted
        by descending connection strength.

        Parameters
        ----------
        region: str, Region
        min_connectivity: float, default: 0
            Regions with connectivity less than this value are discarded.
        max_rows: int, default: None
            Max number of regions with highest connectivity.
        direction: str, default: 'column'
            Choose the direction of profile extraction particularly for
            non-symmetric matrices. ('column' or 'row')
        """
        matrix = self.data
        assert isinstance(matrix, pd.DataFrame)
        matrix_std = self._matrix_std
        if direction.lower() not in ['column', 'row']:
            raise ValueError("Direction can only be 'column' or 'row'")
        if direction.lower() == 'row':
            matrix = matrix.transpose()
            if matrix_std is not None:
                matrix_std = matrix_std.transpose()

        def matches(r1, r2):
            if isinstance(r1, tuple):
                return any(r.matches(r2) for r in r1)
            else:
                assert isinstance(r1, _region.Region)
                return r1.matches(r2)

        # decode region spec
        region_candidates = [r for r in matrix.index if matches(r, region)]
        if len(region_candidates) == 0:
            raise ValueError(f"Invalid region specification: {region}")
        if len(region_candidates) > 1:
            raise ValueError(f"Region specification {region} matched more than one profile: {region_candidates}")
        region = region_candidates[0]

        # create DataFrame
        data = matrix[region].to_frame('mean')
        if matrix_std is not None:
            data = pd.concat([data, matrix_std[region].rename('std')], axis=1)

        last_index = len(data) if max_rows is None else min(max_rows, len(data))

        data = (
            data
            .query(f'`mean` > {min_connectivity}')
            .sort_values(by="mean", ascending=False)
            .rename_axis('Target regions')
        )[:last_index]

        return Tabular(
            description=self.description,
            modality=f"{self.modality} {self.cohort}",
            anchor=_anchor.AnatomicalAnchor(
                species=list(self.anchor.species)[0],
                region=region
            ),
            data=data,
            datasets=self.datasets
        )

    def plot(
        self,
        regions: Union[str, _region.Region, List[Union[str, _region.Region]]] = None,
        min_connectivity: float = 0,
        max_rows: int = None,
        direction: Literal['column', 'row'] = 'column',
        logscale: bool = False,
        *args,
        backend="matplotlib",
        **kwargs
    ):
        """
        Parameters
        ----------
        regions: Union[str, _region.Region], None
            If None, returns the full connectivity matrix.
            If a region is provided, returns the profile for that region.
            If list of regions is provided, returns the matrix for the selected
            regions.
        min_connectivity: float, default 0
            Only for region profile.
        max_rows: int, default None
            Only for region profile.
        direction: 'column' or 'row', default: 'column'
            Only for matrix.
        logscale: bool, default: False
        backend: str, default: "matplotlib" for profiles and "nilearn" for matrices
        """
        if regions is None or isinstance(regions, list):
            plot_matrix_backend = "nilearn" if backend == "matplotlib" else backend
            return self._plot_matrix(
                regions=regions, logscale=logscale, *args,
                backend=plot_matrix_backend,
                **kwargs
            )

        profile = self.get_profile(regions, min_connectivity, max_rows, direction)
        kwargs["kind"] = kwargs.get("kind", "barh")
        if backend == "matplotlib":
            kwargs["logy"] = kwargs.get("logy", logscale)
            return profile.plot(*args, backend=backend, **kwargs)
        elif backend == "plotly":
            kwargs.update({
                "color": kwargs.get("color", profile.data.columns[0]),
                "x": kwargs.get("x", profile.data.columns[0]),
                "y": kwargs.get("y", [r.name for r in profile.data.index]),
                "log_x": logscale,
                "labels": {"y": " ", "x": ""},
                "color_continuous_scale": "jet",
                "width": 600, "height": 15 * len(profile.data)
            })
            fig = profile.data.plot(*args, backend=backend, **kwargs)
            fig.update_layout({
                "font": dict(size=9),
                "yaxis": {"autorange": "reversed"},
                "coloraxis": {"colorbar": {
                    "orientation": "h", "title": "", "xpad": 0, "ypad": 10
                }},
                "margin": dict(l=0, r=0, b=0, t=0, pad=0)
            })
            return fig
        else:
            return profile.data.plot(*args, backend=backend, **kwargs)

    def get_profile_colorscale(
        self,
        region: Union[str, _region.Region],
        min_connectivity: float = 0,
        max_rows: int = None,
        direction: Literal['column', 'row'] = 'column',
        colorgradient: str = "jet"
    ) -> Iterator[Tuple[_region.Region, Tuple[int, int, int]]]:
        """
        Extract the colorscale corresponding to the regional profile from the
        matrix sorted by the values. See `get_profile` for further details.

        Note:
        -----
        Requires `plotly`.

        Parameters
        ----------
        region: str, Region
        min_connectivity: float, default: 0
            Regions with connectivity less than this value are discarded.
        max_rows: int, default: None
            Max number of regions with highest connectivity.
        direction: str, default: 'column'
            Choose the direction of profile extraction particularly for
            non-symmetric matrices. ('column' or 'row')
        colorgradient: str, default: 'jet'
            The gradient used to extract colorscale.
        Returns
        -------
        Iterator[Tuple[_region.Region, Tuple[int, int, int]]]
            Color values are in RGB 255.
        """
        from plotly.express.colors import sample_colorscale
        profile = self.get_profile(region, min_connectivity, max_rows, direction)
        normalized = profile.data / profile.data.max()
        colorscale = sample_colorscale(
            colorgradient,
            normalized.values.reshape(len(profile.data))
        )
        return zip(
            profile.data.index.values,
            [eval(c.removeprefix('rgb')) for c in colorscale]
        )

    def __len__(self):
        return len(self._filename)

    def __str__(self):
        return self.name

    def compute_centroids(self, space):
        """
        Computes the list of centroid coordinates corresponding to
        matrix rows, in the given reference space.

        Parameters
        ----------
        space: Space, str
        Returns
        -------
        list[tuple(float, float, float)]
        """
        result = []
        parcellations = self.anchor.represented_parcellations()
        assert len(parcellations) == 1
        parcmap = next(iter(parcellations)).get_map(space)
        all_centroids = parcmap.compute_centroids()
        for regionname in self.regions:
            region = parcmap.parcellation.get_region(regionname, allow_tuple=True)
            if isinstance(region, tuple):  # deal with sets of matched regions
                found = [c for r in region for c in r if c.name in all_centroids]
            else:
                found = [r for r in region if r.name in all_centroids]
            assert len(found) > 0
            result.append(
                tuple(pointcloud.PointCloud(
                    [all_centroids[r.name] for r in found], space=space
                ).centroid)
            )
        return result

    def _arraylike_to_dataframe(self, array: Union[np.ndarray, pd.DataFrame]) -> pd.DataFrame:
        """
        Convert a numpy array with the connectivity matrix to
        a DataFrame with regions as column and row headers.
        """
        if not isinstance(array, np.ndarray):
            array = array.to_numpy()
        assert array.shape[0] == array.shape[1], f"Connectivity matrices must be square but found {array.shape}"
        if not (array == array.T).all():
            logger.warning("The connectivity matrix is not symmetric.")
        df = pd.DataFrame(array)
        parcellations = self.anchor.represented_parcellations()
        assert len(parcellations) == 1
        parc = next(iter(parcellations))
        with QUIET:
            indexmap = {
                i: parc.get_region(regionname, allow_tuple=True)
                for i, regionname in enumerate(self.regions)
            }
        nrows = array.shape[0]
        try:
            assert len(indexmap) == nrows
            remapper = {
                label - min(indexmap.keys()): region
                for label, region in indexmap.items()
            }
            df = df.rename(index=remapper).rename(columns=remapper)
        except Exception:
            raise RuntimeError("Could not decode connectivity matrix regions.")
        return df

    def _load_matrix(self):
        """
        Extract connectivity matrix.
        """
        if isinstance(self._connector, HttpRequest):
            array = self._connector.data
        else:
            array = self._connector.get(self._filename, decode_func=self._decode_func)
        nrows = array.shape[0]
        if array.shape[1] != nrows:
            raise RuntimeError(
                f"Non-quadratic connectivity matrix {nrows}x{array.shape[1]} "
                f"from {self._filename} in {str(self._connector)}"
            )
        self._matrix = self._arraylike_to_dataframe(array)



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Representation of a brain region."""

import re
from typing import List, Union, Iterable, Dict, Callable, Tuple, Set
from difflib import SequenceMatcher
import json
from functools import wraps, reduce
from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache

import anytree
from ebrains_drive import BucketApiClient

from . import concept, structure, space as _space, parcellation as _parcellation, assignment
from ..retrieval.cache import cache_user_fn
from ..locations import location, pointcloud, boundingbox as _boundingbox
from ..volumes import parcellationmap, volume
from ..commons import (
    logger,
    MapType,
    clear_name,
    InstanceTable,
    QUIET,
)
from ..exceptions import NoMapAvailableError, SpaceWarpingFailedError


REGEX_TYPE = type(re.compile("test"))

THRESHOLD_STATISTICAL_MAPS = None


class Region(anytree.NodeMixin, concept.AtlasConcept, structure.BrainStructure):
    """
    Representation of a region with name and more optional attributes
    """

    _regex_re = re.compile(r'^\/(?P<expression>.+)\/(?P<flags>[a-zA-Z]*)$')
    _accepted_flags = "aiLmsux"

    _GETMASK_CACHE = {}
    _GETMASK_CACHE_MAX_ENTRIES = 1

    def __init__(
        self,
        name: str,
        children: List["Region"] = [],
        parent: "Region" = None,
        shortname: str = "",
        description: str = "",
        modality: str = "",
        publications: list = [],
        datasets: list = [],
        rgb: str = None,
        spec=None,
        prerelease: bool = False,
    ):
        """
        Constructs a new Region object.

        Parameters
        ----------
        name : str
            Human-readable name of the region
        children: list[Region]
        parent: Region
        shortname: str
            Shortform of human-readable name (optional)
        description: str
            Textual description of the parcellation
        modality:  str or None
            Specification of the modality used for specifying this region
        publications: list
            List of associated publications, each a dictionary with "doi"
            and/or "citation" fields
        datasets: list
            datasets associated with this region
        rgb: str, default: None
            Hexcode of preferred color of this region (e.g. "#9FE770")
        spec: dict, default: None
            The preconfigured specification.
        """
        anytree.NodeMixin.__init__(self)
        concept.AtlasConcept.__init__(
            self,
            identifier=None,  # lazy property implementation below
            name=clear_name(name),
            species=None,  # lazy property implementation below
            shortname=shortname,
            description=description,
            modality=modality,
            publications=publications,
            datasets=datasets,
            spec=spec,
            prerelease=prerelease,
        )

        # anytree node will take care to use this appropriately
        self.parent: "Region" = parent
        self.children: List["Region"] = children
        # convert hex to int tuple if rgb is given
        self.rgb = (
            None if rgb is None
            else tuple(int(rgb[p:p + 2], 16) for p in [1, 3, 5])
        )
        self._supported_spaces: Set[_space.Space] = None  # computed on 1st call of self.supported_spaces
        self._str_aliases = None
        self.find = lru_cache(maxsize=3)(self.find)

    def get_related_regions(self) -> Iterable["RegionRelationAssessments"]:
        """
        Get assessments on relations of this region to others defined on EBRAINS.

        Yields
        ------
        assignment.Qualification

        Example
        -------
        >>> region = siibra.get_region("monkey", "PG")
        >>> for assessment in region.get_related_regions():
        >>>    print(assessment)
        'PG' is homologous to 'Area PGa (IPL)'
        'PG' is homologous to 'Area PGa (IPL) left'
        'PG' is homologous to 'Area PGa (IPL) right'
        'PG' is homologous to 'Area PGa (IPL)'
        'PG' is homologous to 'Area PGa (IPL) left'
        'PG' is homologous to 'Area PGa (IPL) right'
        'PG' is homologous to 'Area PGa (IPL)'
        'PG' is homologous to 'Area PGa (IPL) right'
        'PG' is homologous to 'Area PGa (IPL) left'
        """
        yield from RegionRelationAssessments.parse_from_region(self)

    @property
    def id(self):
        if self.parent is None:
            return self.key
        else:
            return f"{self.root.key}_{self.key}"

    @property
    def parcellation(self):
        if isinstance(self.root, _parcellation.Parcellation):
            return self.root
        else:
            return None

    @property
    def species(self):
        # lazy request of the root parcellation's species
        if self._species_cached is None:
            self._species_cached = self.parcellation.species
        return self._species_cached

    @staticmethod
    def copy(other: 'Region'):
        """
        copy constructor must detach the parent to avoid problems with
        the Anytree implementation.
        """
        # create an isolated object, detached from the other's tree
        region = Region(
            name=other.name,
            children=[Region.copy(c) for c in other.children],
            parent=None,
            shortname=other.shortname,
            description=other.description,
            modality=other.modality,
            publications=other.publications,
            datasets=other.datasets,
            rgb=other.rgb)

        for c in region.children:
            c.parent = region
        return region

    @property
    def names(self):
        return {r.name for r in self}

    def __eq__(self, other):
        """
        Compare this region with other objects. If other is a string,
        compare to key, name or id.
        """
        if isinstance(other, Region):
            return self.id == other.id
        if isinstance(other, str):
            if not self._str_aliases:
                self._str_aliases = {
                    self.name,
                    self.key,
                    self.id,
                }
                if self._spec:
                    ebrain_ids = [
                        value
                        for value in self._spec.get("ebrains", {}).values()
                        if isinstance(value, str)
                    ]
                    ebrain_nested_ids = [
                        _id
                        for value in self._spec.get("ebrains", {}).values() if isinstance(value, list)
                        for _id in value
                    ]
                    assert all(isinstance(_id, str) for _id in ebrain_nested_ids)
                    all_ebrain_ids = [
                        *ebrain_ids,
                        *ebrain_nested_ids
                    ]

                    self._str_aliases.update(all_ebrain_ids)

            return other in self._str_aliases
        return False

    def __hash__(self):
        return hash(self.id)

    def has_parent(self, parent):
        return parent in [a for a in self.ancestors]

    def includes(self, region):
        """
        Determine whether this region-tree includes the given region.

        Parameters
        ----------
        region: Region

        Returns
        -------
            bool
                True if the region is in the region-tree.
        """
        return region == self or region in self.descendants

    def find(
        self,
        regionspec,
        filter_children=False,
        find_topmost=True,
    ) -> List['Region']:
        """
        Find regions that match the given region specification in the subtree
        headed by this region.

        Parameters
        ----------
        regionspec: str, regex, Region
            - a string with a possibly inexact name (matched both against the name and the identifier key)
            - a string in '/pattern/flags' format to use regex search (acceptable flags: aiLmsux, see at https://docs.python.org/3/library/re.html#flags)
            - a regex applied to region names
            - a Region object
        filter_children : bool, default: False
            If True, children of matched parents will not be returned
        find_topmost : bool, default: True
            If True (requires `filter_children=True`), will return parent
            structures if all children are matched, even though the parent
            itself might not match the specification.

        Returns
        -------
        list[Region]
            list of regions matching to the regionspec

        Tip
        ---
        See example 01-003, find regions.
        """
        if isinstance(regionspec, str):
            # convert the specified string into a regex for matching
            regex_match = self._regex_re.match(regionspec)
            if regex_match:
                flags = regex_match.group('flags')
                expression = regex_match.group('expression')

                for flag in flags or []:  # catch if flags is nullish
                    if flag not in self._accepted_flags:
                        raise Exception(f"only accepted flag are in {self._accepted_flags}. {flag} is not within them")
                search_regex = (f"(?{flags})" if flags else "") + expression
                regionspec = re.compile(search_regex)

        candidates = list(
            anytree.search.findall(self, lambda node: node.matches(regionspec))
        )

        if len(candidates) > 1 and filter_children:
            filtered = []
            for region in candidates:
                children_included = [c for c in region.children if c in candidates]
                if len(children_included) > 0:
                    filtered.append(region)
                else:
                    if region.parent not in candidates:
                        filtered.append(region)

            # find any non-matched regions of which all children are matched
            if find_topmost:
                complete_parents = list(
                    {
                        r.parent
                        for r in filtered
                        if (r.parent is not None)
                        and all((c in filtered) for c in r.parent.children)
                    }
                )

                if len(complete_parents) == 0:
                    candidates = filtered
                else:
                    # filter child regions again
                    filtered += complete_parents
                    candidates = [
                        r for r in filtered
                        if (r.parent not in filtered) or r == regionspec
                    ]
            else:
                candidates = filtered

        # ensure the result is a list
        if candidates is None:
            candidates = []
        elif isinstance(candidates, Region):
            candidates = [candidates]
        else:
            candidates = list(candidates)

        found_regions = sorted(set(candidates), key=lambda r: r.depth)

        # reverse is set to True, since SequenceMatcher().ratio(), higher == better
        return (
            sorted(
                found_regions,
                reverse=True,
                key=lambda region: SequenceMatcher(None, str(region), regionspec).ratio(),
            )
            if isinstance(regionspec, str) else found_regions
        )

    def matches(self, regionspec):
        """
        Checks whether this region matches the given region specification.

        Parameters
        ----------
        regionspec: str, regex, Region
            - a string with a possibly inexact name, which is matched both against the name and the identifier key,
            - a regex applied to region names,
            - a region object

        Returns
        -------
        bool
            If the regionspec matches to the Region.
        """
        if regionspec not in self._CACHED_MATCHES:
            def splitstr(s):
                return [w for w in re.split(r"[^a-zA-Z0-9.\-]", s) if len(w) > 0]

            if regionspec is None:
                self._CACHED_MATCHES[regionspec] = False

            elif isinstance(regionspec, Region):
                self._CACHED_MATCHES[regionspec] = self == regionspec

            elif isinstance(regionspec, str):
                # string is given, perform lazy string matching
                q = regionspec.lower().strip()
                if q == self.key.lower().strip():
                    self._CACHED_MATCHES[regionspec] = True
                elif q == self.id.lower().strip():
                    self._CACHED_MATCHES[regionspec] = True
                elif q == self.name.lower().strip():
                    self._CACHED_MATCHES[regionspec] = True
                else:
                    # match if all words of the query are also included in the region name
                    W = splitstr(clear_name(self.name.lower()))
                    Q = splitstr(clear_name(regionspec))
                    self._CACHED_MATCHES[regionspec] = all([any(
                        q.lower() == w or 'v' + q.lower() == w
                        for w in W
                    ) for q in Q])

            # TODO since dropping 3.6 support, maybe reimplement as re.Pattern ?
            elif isinstance(regionspec, REGEX_TYPE):
                # match regular expression
                self._CACHED_MATCHES[regionspec] = any(regionspec.search(s) is not None for s in [self.name, self.key])

            elif isinstance(regionspec, (list, tuple)):
                self._CACHED_MATCHES[regionspec] = any(self.matches(_) for _ in regionspec)

            else:
                raise TypeError(
                    f"Cannot interpret region specification of type '{type(regionspec)}'"
                )

        return self._CACHED_MATCHES[regionspec]

    def get_regional_mask(
        self,
        space: Union[str, _space.Space],
        maptype: MapType = MapType.LABELLED,
        threshold: float = 0.0,
    ) -> Union[volume.FilteredVolume, volume.ReducedVolume]:
        """
        Get a binary mask of this region in the given space,
        using the specified MapTypes.

        Parameters
        ----------
        space: Space or str
            The requested reference space
        maptype: MapType, default: SIIBRA_DEFAULT_MAPTYPE
            The type of map to be used ('labelled' or 'statistical')
        threshold: float, default: 0.0
            When fetching a statistical map, use this threshold to convert
            it to a binary mask.

        Returns
        -------
        Volume (use fetch() to get a NiftiImage)
        """
        if isinstance(maptype, str):
            maptype = MapType[maptype.upper()]

        threshold_info = "" if maptype == MapType.LABELLED else f"(threshold: {threshold}) "
        # check cache
        getmask_hash = hash(f"{self.id} - {space} - {maptype}{threshold_info}")
        if getmask_hash in self._GETMASK_CACHE:
            return self._GETMASK_CACHE[getmask_hash]

        name = f"Mask {threshold_info}of '{self.name} ({self.parcellation})' in "
        try:
            regional_map = self.get_regional_map(space=space, maptype=maptype)
            if maptype == MapType.LABELLED:
                assert threshold == 0.0, f"threshold can only be set for {MapType.STATISTICAL} maps."
                result = volume.FilteredVolume(
                    parent_volume=regional_map,
                    label=regional_map.label,
                    fragment=regional_map.fragment
                )
                result._boundingbox = None
            if maptype == MapType.STATISTICAL:
                result = volume.FilteredVolume(
                    parent_volume=regional_map,
                    threshold=threshold
                )
                if threshold == 0.0:
                    result._boundingbox = regional_map._boundingbox
            name += f"'{result.space}'"
        except NoMapAvailableError as e:
            # This region is not mapped directly in any map in the registry.
            # Try building a map from the child regions
            if (len(self.children) > 0) and self.mapped_in_space(space, recurse=True):
                mapped_descendants: List[Region] = [
                    d for d in self.descendants if d.mapped_in_space(space, recurse=False)
                ]
                logger.info(f"{self.name} is not mapped in {space}. Merging the masks of its {len(mapped_descendants)} map descendants.")
                descendant_volumes = [
                    descendant.get_regional_mask(space=space, maptype=maptype, threshold=threshold)
                    for descendant in mapped_descendants
                ]
                with QUIET:
                    result = volume.ReducedVolume(
                        descendant_volumes,
                        new_labels=[1] * len(descendant_volumes)
                    )
                name += f"'{result.space}' (built by merging the mask {threshold_info}of its descendants)"
            else:
                raise e
        result._name = name

        while len(self._GETMASK_CACHE) > self._GETMASK_CACHE_MAX_ENTRIES:
            self._GETMASK_CACHE.pop(next(iter(self._GETMASK_CACHE)))
        self._GETMASK_CACHE[getmask_hash] = result
        return result

    def get_regional_map(
        self,
        space: Union[str, _space.Space],
        maptype: MapType = MapType.LABELLED,
    ) -> Union[volume.FilteredVolume, volume.Volume, volume.Subvolume]:
        """
        Get a volume representing this region in the given space and MapType.

        Note
        ----
        If a region is not mapped in any of the `Map`s in the registry, then
        siibra will get the maps of its children recursively and merge them.
        If no map is available this way as well, an exception is raised.

        Parameters
        ----------
        space: Space or str
            The requested reference space
        maptype: MapType, default: SIIBRA_DEFAULT_MAPTYPE
            The type of map to be used ('labelled' or 'statistical')

        Returns
        -------
        Volume (use fetch() to get a NiftiImage)
        """
        if isinstance(maptype, str):
            maptype = MapType[maptype.upper()]

        # prepare space instance
        if isinstance(space, str):
            space = _space.Space.get_instance(space)

        # see if we find a map supporting the requested region
        for m in parcellationmap.Map.registry():
            if (
                m.space.matches(space)
                and m.parcellation == self.parcellation
                and m.provides_image
                and m.maptype == maptype
                and self.name in m.regions
            ):
                return m.get_volume(region=self)
        raise NoMapAvailableError(
            f"{self.name} is not mapped in {space} as a {str(maptype)} map."
            " Please try getting the children or getting the mask."
        )

    def mapped_in_space(self, space, recurse: bool = False) -> bool:
        """
        Verifies whether this region is defined by an explicit map in the given space.

        Parameters
        ----------
        space: Space or str
            reference space
        recurse: bool, default: False
            If True, check if itself or all child regions are mapped instead recursively.
        Returns
        -------
        bool
        """
        from ..volumes.parcellationmap import Map
        for m in Map.registry():
            if (
                m.space.matches(space)
                and m.parcellation.matches(self.parcellation)
                and self.name in m.regions
            ):
                return True
        if recurse and not self.is_leaf:
            # check if all children are mapped instead
            return all(c.mapped_in_space(space, recurse=True) for c in self.children)
        return False

    @property
    def supported_spaces(self) -> List[_space.Space]:
        """
        The list of spaces for which a mask could be extracted from an existing
        map or combination of masks of its children.
        """
        if self._supported_spaces is None:
            self._supported_spaces = sorted({
                s for s in _space.Space.registry()
                if self.mapped_in_space(s, recurse=True)
            })
        return self._supported_spaces

    @property
    def spaces(self):
        return InstanceTable(
            matchfunc=_space.Space.matches,
            elements={s.key: s for s in self.supported_spaces},
        )

    def __contains__(self, other: Union[location.Location, 'Region']) -> bool:
        if isinstance(other, Region):
            return len(self.find(other)) > 0
        else:
            try:
                regionmap = self.get_regional_mask(space=other.space)
                return regionmap.__contains__(other)
            except NoMapAvailableError:
                return False

    def assign(self, other: structure.BrainStructure) -> assignment.AnatomicalAssignment:
        """
        Compute assignment of a location to this region.

        Two cases:
        1) other is location -> get region map, call regionmap.assign(other)
        2) other is region -> just do a semantic check for the regions

        Parameters
        ----------
        other : Location or Region

        Returns
        -------
        assignment.AnatomicalAssignment or None
            None if there is no assignment.Qualification found.
        """
        if (self, other) in self._ASSIGNMENT_CACHE:
            return self._ASSIGNMENT_CACHE[self, other]
        if (other, self) in self._ASSIGNMENT_CACHE:
            if self._ASSIGNMENT_CACHE[other, self] is None:
                return None
            return self._ASSIGNMENT_CACHE[other, self].invert()

        if isinstance(other, (location.Location, volume.Volume)):
            if self.mapped_in_space(other.space, recurse=True):
                regionmap = self.get_regional_mask(other.space)
                self._ASSIGNMENT_CACHE[self, other] = regionmap.assign(other)
                return self._ASSIGNMENT_CACHE[self, other]

            if isinstance(other, _boundingbox.BoundingBox):  # volume.intersection(bbox) gets boundingbox anyway
                try:
                    regionbbox_otherspace = self.get_boundingbox(other.space, restrict_space=False)
                    if regionbbox_otherspace is not None:
                        self._ASSIGNMENT_CACHE[self, other] = regionbbox_otherspace.assign(other)
                        return self._ASSIGNMENT_CACHE[self, other]
                except Exception as e:
                    logger.debug(e)

            assignment_result = None
            for targetspace in self.supported_spaces:
                try:
                    other_warped = other.warp(targetspace)
                    regionmap = self.get_regional_mask(targetspace)
                    assignment_result = regionmap.assign(other_warped)
                except SpaceWarpingFailedError:
                    try:
                        regionbbox_targetspace = self.get_boundingbox(
                            targetspace, restrict_space=True
                        )
                        if regionbbox_targetspace is None:
                            continue
                        regionbbox_warped = regionbbox_targetspace.warp(other.space)
                    except SpaceWarpingFailedError:
                        continue
                    assignment_result = regionbbox_warped.assign(other)
                except Exception as e:
                    logger.debug(e)
                    continue
                break
            self._ASSIGNMENT_CACHE[self, other] = assignment_result
        else:  # other is a Region
            assert isinstance(other, Region)
            if self == other:
                qualification = assignment.Qualification.EXACT
            elif self.__contains__(other):
                qualification = assignment.Qualification.CONTAINS
            elif other.__contains__(self):
                qualification = assignment.Qualification.CONTAINED
            else:
                qualification = None
            if qualification is None:
                self._ASSIGNMENT_CACHE[self, other] = None
            else:
                self._ASSIGNMENT_CACHE[self, other] = assignment.AnatomicalAssignment(self, other, qualification)
        return self._ASSIGNMENT_CACHE[self, other]

    def tree2str(self):
        """Render region-tree as a string"""
        return "\n".join(
            "%s%s" % (pre, node.name)
            for pre, _, node
            in anytree.RenderTree(self, style=anytree.render.ContRoundStyle)
        )

    def render_tree(self):
        """Prints the tree representation of the region"""
        print(self.tree2str())

    def get_boundingbox(
        self,
        space: _space.Space,
        maptype: MapType = MapType.LABELLED,
        threshold_statistical: float = 0.0,
        restrict_space: bool = True,
        **fetch_kwargs
    ) -> Union[_boundingbox.BoundingBox, None]:
        """
        Compute the bounding box of this region in the given space.

        Parameters
        ----------
        space: Space or str
            Requested reference space
        maptype: MapType, default: MapType.LABELLED
            Type of map to build ('labelled' will result in a binary mask,
            'statistical' attempts to build a statistical mask, possibly by
            elementwise maximum of statistical maps of children)
        threshold_statistical: float, default: 0.0
            When masking a statistical map, use this threshold to convert
            it to a binary mask before finding its bounding box.
        restrict_space: bool, default: False
            If True, it will not try to fetch maps from other spaces and warp
            its boundingbox to requested space.

        Returns
        -------
        BoundingBox
        """
        spaceobj = _space.Space.get_instance(space)
        try:
            mask = self.get_regional_mask(
                spaceobj, maptype=maptype, threshold=threshold_statistical
            )
            return mask.get_boundingbox(
                clip=True,
                background=0.0,
                **fetch_kwargs
            )
        except (RuntimeError, ValueError):
            if restrict_space:
                return None
            for other_space in self.parcellation.spaces - spaceobj:
                try:
                    mask = self.get_regional_mask(
                        other_space,
                        maptype=maptype,
                        threshold=threshold_statistical,
                    )
                    bbox = mask.get_boundingbox(clip=True, background=0.0, **fetch_kwargs)
                    if bbox is not None:
                        try:
                            bbox_warped = bbox.warp(spaceobj)
                        except SpaceWarpingFailedError:
                            continue
                        logger.debug(
                            f"No bounding box for {self.name} defined in {spaceobj.name}, "
                            f"warped the bounding box from {other_space.name} instead."
                        )
                        return bbox_warped
                except RuntimeError:
                    continue
        logger.error(f"Could not compute bounding box for {self.name}.")
        return None

    def compute_centroids(
        self,
        space: _space.Space,
        maptype: MapType = MapType.LABELLED,
        threshold_statistical: float = 0.0,
        split_components: bool = True,
        **fetch_kwargs,
    ) -> pointcloud.PointCloud:
        """
        Compute the centroids of the region in the given space.

        Parameters
        ----------
        space: Space
            reference space in which the computation will be performed
        maptype: MapType, default: MapType.LABELLED
            Type of map to build ('labelled' will result in a binary mask,
            'statistical' attempts to build a statistical mask, possibly by
            elementwise maximum of statistical maps of children)
        threshold_statistical: float, default: 0.0
            When masking a statistical map, use this threshold to convert
            it to a binary mask before finding its centroids.

        Returns
        -------
        PointCloud
            Found centroids (as Point objects) in a PointCloud

        Note
        ----
        A region can generally have multiple centroids if it has multiple
        connected components in the map.
        """
        props = self.spatial_props(
            space=space,
            maptype=maptype,
            threshold_statistical=threshold_statistical,
            split_components=split_components,
            **fetch_kwargs,
        )
        return pointcloud.PointCloud(
            [c.centroid for c in props],
            space=space
        )

    def spatial_props(
        self,
        space: _space.Space,
        maptype: MapType = MapType.LABELLED,
        threshold_statistical: float = 0.0,
        split_components: bool = True,
        **fetch_kwargs,
    ):
        """
        Compute spatial properties for connected components of this region in the given space.

        Parameters
        ----------
        space: Space
            reference space in which the computation will be performed
        maptype: MapType, default: MapType.LABELLED
            Type of map to build ('labelled' will result in a binary mask,
            'statistical' attempts to build a statistical mask, possibly by
            elementwise maximum of statistical maps of children)
        threshold_statistical: float, default: 0.0
            if not None, masks will be preferably constructed by thresholding
            statistical maps with the given value.

        Returns
        -------
        List
            List of region's component spatial properties
        """
        if not isinstance(space, _space.Space):
            space = _space.Space.get_instance(space)

        # build binary mask of the image
        try:
            region_vol = self.get_regional_mask(
                space, maptype=maptype, threshold=threshold_statistical
            )
        except NoMapAvailableError:
            raise ValueError(
                f"Spatial properties of {self.name} cannot be computed in {space.name}. "
                "This region is only mapped in these spaces: "
                f"{', '.join(s.name for s in self.supported_spaces)}"
            )

        return region_vol.compute_spatial_props(
            split_components=split_components, **fetch_kwargs
        )

    def __iter__(self):
        """
        Returns an iterator that goes through all regions in this subtree
        (including this parent region)
        """
        return anytree.PreOrderIter(self)

    def intersection(self, other: "location.Location") -> "location.Location":
        """Use this region for filtering a location object."""

        if self.mapped_in_space(other.space, recurse=True):
            try:
                volume = self.get_regional_mask(other.space)
                if volume is not None:
                    return volume.intersection(other)
            except NotImplementedError:
                intersections = [child.intersection(other) for child in self.children]
                return reduce(lambda a, b: a.union(b), intersections)

        for space in self.supported_spaces:
            if space.provides_image:
                try:
                    volume = self.get_regional_mask(space)
                    if volume is not None:
                        intersection = volume.intersection(other)
                        logger.info(f"Warped {other} to {space} to find the intersection.")
                        return intersection
                except SpaceWarpingFailedError:
                    continue

        return None


@cache_user_fn
def _get_related_regions_str(pe_id: str) -> Tuple[Tuple[str, str, str, str], ...]:
    logger.info("LONG CALC...", pe_id)
    return_val = []
    region_relation_assessments = RegionRelationAssessments.translate_pes(pe_id, pe_id)
    for asgmt in region_relation_assessments:
        assert isinstance(asgmt, RegionRelationAssessments), f"Expecting type to be of RegionRelationAssessments, but is {type(asgmt)}"
        assert isinstance(asgmt.assigned_structure, Region), f"Expecting assigned structure to be of type Region, but is {type(asgmt.assigned_structure)}"
        return_val.append((
            asgmt.assigned_structure.parcellation.id,
            asgmt.assigned_structure.name,
            asgmt.qualification.name,
            asgmt.explanation
        ))
    return tuple(return_val)


def get_peid_from_region(region: Region) -> str:
    """
    Given a region, obtain the Parcellation Entity ID.

    Parameters
    ----------
    region : Region

    Returns
    -------
    str
    """
    if region._spec:
        region_peid = region._spec.get("ebrains", {}).get("openminds/ParcellationEntity")
        if region_peid:
            return region_peid
    # In some cases (e.g. Julich Brain, PE is defined on the parent leaf nodes)
    if region.parent and region.parent._spec:
        parent_peid = region.parent._spec.get("ebrains", {}).get("openminds/ParcellationEntity")
        if parent_peid:
            return parent_peid
    return None


def get_related_regions(region: Region) -> Iterable["RegionRelationAssessments"]:
    """
    Get assessments on relations of a region to others defined on EBRAINS.

    Parameters
    ----------
    region: Region

    Yields
    ------
    assignment.Qualification

    Example
    -------
    >>> region = siibra.get_region("monkey", "PG")
    >>> for assessment in siibra.core.region.get_related_regions(region):
    >>>    print(assessment)
    'PG' is homologous to 'Area PGa (IPL)'
    'PG' is homologous to 'Area PGa (IPL) left'
    'PG' is homologous to 'Area PGa (IPL) right'
    'PG' is homologous to 'Area PGa (IPL)'
    'PG' is homologous to 'Area PGa (IPL) left'
    'PG' is homologous to 'Area PGa (IPL) right'
    'PG' is homologous to 'Area PGa (IPL)'
    'PG' is homologous to 'Area PGa (IPL) right'
    'PG' is homologous to 'Area PGa (IPL) left'
    """
    logger.info("get related region called")
    pe_id = get_peid_from_region(region)
    if not pe_id:
        return []

    for parc_id, region_name, qual, explanation in _get_related_regions_str(pe_id):
        parc = _parcellation.Parcellation.get_instance(parc_id)
        found_region = parc.get_region(region_name)
        yield RegionRelationAssessments(region, found_region, qual, explanation)


_get_reg_relation_asmgt_types: Dict[str, Callable] = {}


def _register_region_reference_type(ebrain_type: str):
    def outer(fn: Callable):
        _get_reg_relation_asmgt_types[ebrain_type] = fn

        @wraps(fn)
        def inner(*args, **kwargs):
            return fn(*args, **kwargs)
        return inner
    return outer


class RegionRelationAssessments(assignment.AnatomicalAssignment[Region]):
    """
    A collection of methods on finding related regions and the quantification
    of the relationship.
    """

    anony_client = BucketApiClient()

    @staticmethod
    def get_uuid(long_id: Union[str, Dict]) -> str:
        """
        Returns the uuid portion of either a fully formed openminds id, or get
        the 'id' property first, and extract the uuid portion of the id.

        Parameters
        ----------
        long_id: str, dict[str, str]

        Returns
        -------
        str

        Raises
        ------
        AssertionError
        RuntimeError
        """
        if isinstance(long_id, str):
            pass
        elif isinstance(long_id, dict):
            long_id = long_id.get("id")
            assert isinstance(long_id, str)
        else:
            raise RuntimeError("uuid arg must be str or object")
        uuid_search = re.search(r"(?P<uuid>[a-f0-9-]+)$", long_id)
        assert uuid_search, "uuid not found"
        return uuid_search.group("uuid")

    @staticmethod
    def parse_id_arg(_id: Union[str, List[str]]) -> List[str]:
        """
        Normalizes the ebrains id property. The ebrains id field can be either
        a str or list[str]. This method normalizes it to always be list[str].

        Parameters
        ----------
        _id: strl, list[str]

        Returns
        -------
        list[str]

        Raises
        ------
        RuntimeError
        """
        if isinstance(_id, list):
            assert all(isinstance(_i, str) for _i in _id), "all instances of pev should be str"
        elif isinstance(_id, str):
            _id = [_id]
        else:
            raise RuntimeError("parse_pev error: arg must be either list of str or str")
        return _id

    @classmethod
    def get_object(cls, obj: str):
        """
        Gets given a object (path), loads the content and serializes to json.
        Relative to the bucket 'reference-atlas-data'.

        Parameters
        ----------
        obj: str

        Returns
        -------
        dict
        """
        bucket = cls.anony_client.buckets.get_bucket("reference-atlas-data")
        return json.loads(bucket.get_file(obj).get_content())

    @classmethod
    def get_snapshot_factory(cls, type_str: str):
        """
        Factory method for given type.

        Parameters
        ----------
        type_str: str

        Returns
        -------
        Callable[[str|list[str]], dict]
        """
        def get_objects(_id: Union[str, List[str]]):
            _id = cls.parse_id_arg(_id)
            with ThreadPoolExecutor() as ex:
                return list(
                    ex.map(
                        cls.get_object,
                        [f"ebrainsquery/v3/{type_str}/{_}.json" for _ in _id]
                    ))
        return get_objects

    @classmethod
    def parse_relationship_assessment(cls, src: "Region", assessment):
        """
        Given a region, and the fetched assessment json, yield
        RegionRelationAssignment object.

        Parameters
        ----------
        src: Region
        assessment: dict

        Returns
        -------
        Iterable[RegionRelationAssessments]
        """
        all_regions = [
            region
            for p in _parcellation.Parcellation.registry()
            for region in p
        ]

        overlap = assessment.get("qualitativeOverlap")
        targets = assessment.get("relationAssessment") or assessment.get("inRelationTo")
        assert len(overlap) == 1, f"should be 1&o1 overlap {len(overlap)!r} "
        overlap, = overlap
        for target in targets:
            target_id = cls.get_uuid(target)

            found_targets = [
                region
                for region in all_regions
                if region == target_id
            ]

            for found_target in found_targets:
                yield cls(
                    query_structure=src,
                    assigned_structure=found_target,
                    qualification=assignment.Qualification.parse_relation_assessment(overlap)
                )

            if "https://openminds.ebrains.eu/sands/ParcellationEntity" in target.get("type"):
                pev_uuids = [
                    cls.get_uuid(has_version)
                    for pe in cls.get_snapshot_factory("ParcellationEntity")(target_id)
                    for has_version in pe.get("hasVersion")
                ]
                for reg in all_regions:
                    if reg in pev_uuids:
                        yield cls(
                            query_structure=src,
                            assigned_structure=reg,
                            qualification=assignment.Qualification.parse_relation_assessment(overlap)
                        )

    @classmethod
    @_register_region_reference_type("openminds/CustomAnatomicalEntity")
    def translate_cae(cls, src: "Region", _id: Union[str, List[str]]):
        """Register how CustomAnatomicalEntity should be parsed

        Parameters
        ----------
        src: Region
        _id: str|list[str]

        Returns
        -------
        Iterable[RegionRelationAssessments]
        """
        caes = cls.get_snapshot_factory("CustomAnatomicalEntity")(_id)
        for cae in caes:
            for ass in cae.get("relationAssessment", []):
                yield from cls.parse_relationship_assessment(src, ass)

    @classmethod
    @_register_region_reference_type("openminds/ParcellationEntity")
    def translate_pes(cls, src: "Region", _id: Union[str, List[str]]):
        """
        Register how ParcellationEntity should be parsed

        Parameters
        ----------
        src: Region
        _id: str|list[str]

        Returns
        -------
        Iterable[RegionRelationAssessments]
        """
        pes = cls.get_snapshot_factory("ParcellationEntity")(_id)

        all_regions = [
            region
            for p in _parcellation.Parcellation.registry()
            for region in p
        ]

        for pe in pes:
            for region in all_regions:
                if region is src:
                    continue
                region_peid = get_peid_from_region(region)
                if region_peid and (region_peid in pe.get("id")):
                    yield cls(
                        query_structure=src,
                        assigned_structure=region,
                        qualification=assignment.Qualification.OTHER_VERSION
                    )

            # homologuous
            relations = pe.get("inRelationTo", [])
            for relation in relations:
                yield from cls.parse_relationship_assessment(src, relation)

    @classmethod
    @_register_region_reference_type("openminds/ParcellationEntityVersion")
    def translate_pevs(cls, src: "Region", _id: Union[str, List[str]]):
        """
        Register how ParcellationEntityVersion should be parsed

        Parameters
        ----------
        src: Region
        _id: str|list[str]

        Returns
        -------
        Iterable[RegionRelationAssessments]
        """
        pe_uuids = [
            uuid for uuid in
            {
                cls.get_uuid(pe)
                for pev in cls.get_snapshot_factory("ParcellationEntityVersion")(_id)
                for pe in pev.get("isVersionOf")
            }
        ]
        yield from cls.translate_pes(src, pe_uuids)

    @classmethod
    def parse_from_region(cls, region: "Region") -> Iterable["RegionRelationAssessments"]:
        """
        Main entry on how related regions should be retrieved. Given a region,
        retrieves all RegionRelationAssessments

        Parameters
        ----------
        region: Region

        Returns
        -------
        Iterable[RegionRelationAssessments]
        """
        if not region._spec:
            return None
        for ebrain_type, ebrain_ref in region._spec.get("ebrains", {}).items():
            if ebrain_type in _get_reg_relation_asmgt_types:
                fn = _get_reg_relation_asmgt_types[ebrain_type]
                yield from fn(cls, region, ebrain_ref)



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handles the relation between study targets and BrainStructures."""

from typing import Union, List, Dict, Iterable

from ..commons import Species, logger
from ..core.structure import BrainStructure
from ..core.assignment import AnatomicalAssignment, Qualification
from ..locations.location import Location
from ..core.parcellation import Parcellation, find_regions
from ..core.region import Region
from ..core.space import Space
from ..exceptions import SpaceWarpingFailedError
from ..vocabularies import REGION_ALIASES


class AnatomicalAnchor:
    """
    Anatomical anchor to an atlas region, a geometric primitive in an atlas
    reference space, or both.
    """

    _MATCH_MEMO: Dict[str, Dict[Region, Qualification]] = {}

    def __init__(
        self,
        species: Union[List[Species], Species, str],
        location: Location = None,
        region: Union[str, Region] = None
    ):

        if isinstance(species, (str, Species)):
            self.species = {Species.decode(species)}
        elif isinstance(species, Iterable):
            assert all(isinstance(_, Species) for _ in species)
            self.species = set(species)
        else:
            sp = Species.decode(species)
            if sp is None:
                raise ValueError(f"Invalid species specification: {species}")
            else:
                self.species = {sp}
        self._location_cached = location
        self._assignments: Dict[BrainStructure, List[AnatomicalAssignment]] = {}
        self._last_matched_concept = None
        if isinstance(region, dict):
            self._regions_cached = region
            self._regionspec = ", ".join({r.name for r in region.keys()})
        else:
            self._regions_cached = None
            self._regionspec = None
            if isinstance(region, Region):
                self._regions_cached = {region: Qualification.EXACT}
            elif isinstance(region, str):
                # we will decode regions only when needed, see self.regions property
                self._regionspec = region
            else:
                if region is not None:
                    raise ValueError(f"Invalid region specification: {region}")
        self._aliases_cached = None

    @property
    def location(self) -> Location:
        # allow to overwrite in derived classes
        return self._location_cached

    @property
    def parcellations(self) -> List[Parcellation]:
        """
        Return any parcellation objects that regions of this anchor belong to.
        """
        return list({region.root for region in self.regions})

    @property
    def space(self) -> Space:
        # may be overridden by derived classes, e.g. in features.VolumeOfInterest
        return None if self.location is None else self.location.space

    @property
    def region_aliases(self):
        # return any predefined aliases for the region specified in this anchor.
        if self._aliases_cached is None:
            self._aliases_cached: Dict[str, Dict[str, str]] = {
                Species.decode(species_str): region_alias_mapping
                for s in self.species
                for species_str, region_alias_mapping in REGION_ALIASES.get(str(s), {}).get(self._regionspec, {}).items()
            }
        return self._aliases_cached

    @property
    def has_region_aliases(self) -> bool:
        return len(self.region_aliases) > 0

    @property
    def regions(self) -> Dict[Region, Qualification]:
        """
        Return the list of regions associated with this anchor.
        Decode the self._regionspec string into region objects now,
        if applicable and called for the first time.
        """
        # decoding region strings is quite compute intensive, so we cache this at the class level
        if self._regions_cached is not None:
            return self._regions_cached

        if self._regionspec is None:
            self._regions_cached = dict()
            return self._regions_cached

        match_key = self._regionspec + '-' + str(self.species)
        if match_key not in self.__class__._MATCH_MEMO:
            # decode the region specification into a dict of region objects and assignment qualifications
            regions = {
                region: Qualification.EXACT
                for region in find_regions(self._regionspec, filter_children=True, find_topmost=False)
                if region.species in self.species
            }
            # add more regions from possible aliases of the region spec
            for alt_species, aliases in self.region_aliases.items():
                for alias_regionspec, qualificationspec in aliases.items():
                    for r in find_regions(alias_regionspec, filter_children=True, find_topmost=False):
                        if r.species != alt_species:
                            continue
                        if r not in regions:
                            regions[r] = Qualification[qualificationspec.upper()]

            self.__class__._MATCH_MEMO[match_key] = regions
        self._regions_cached = self.__class__._MATCH_MEMO[match_key]

        return self._regions_cached

    def __str__(self):
        parcs = {p.id: p.name for p in self.represented_parcellations()}
        if len(parcs) == 1 and self._regionspec in [pid for pid in parcs]:
            region = parcs[self._regionspec]  # if parcellation was anchored with the id instead of the name
        else:
            region = "" if self._regionspec is None else str(self._regionspec)
        location = "" if self.location is None else str(self.location)
        separator = " " if min(len(region), len(location)) > 0 else ""
        if region and location:
            return region + " with " + location
        else:
            return region + separator + location

    def assign(self, concept: Union[BrainStructure, Space]) -> AnatomicalAssignment:
        """
        Match this anchor to a query concept. Assignments are cached at runtime,
        so repeated assignment with the same concept will be cheap.
        """
        if isinstance(concept, Space):
            if self.location is not None and self.location.space.matches(concept):
                return [AnatomicalAssignment(concept, self.location, Qualification.CONTAINED)]
            else:
                return []

        if concept not in self._assignments:
            assignments: List[AnatomicalAssignment] = []
            if self.location is not None:
                try:
                    assignments.append(self.location.assign(concept))
                except SpaceWarpingFailedError as e:
                    logger.debug(e)
            for region in self.regions:
                assignments.append(region.assign(concept))
            self._assignments[concept] = sorted(a for a in assignments if a is not None)

        self._last_matched_concept = concept if len(self._assignments[concept]) > 0 else None
        return self._assignments[concept]

    def matches(self, concept: Union[BrainStructure, Space]) -> bool:
        return len(self.assign(concept)) > 0

    def represented_parcellations(self) -> List[Parcellation]:
        """
        Return any parcellation objects that this anchor explicitly points to.
        """
        return [r for r in self.regions if isinstance(r, Parcellation)]

    @property
    def last_match_result(self) -> List[AnatomicalAssignment]:
        return self._assignments.get(self._last_matched_concept, [])

    @property
    def last_match_description(self) -> str:
        if self.last_match_result is None:
            return ""
        else:
            return ' and '.join({str(_) for _ in self.last_match_result})

    def __add__(self, other: 'AnatomicalAnchor') -> 'AnatomicalAnchor':
        if not isinstance(other, AnatomicalAnchor):
            raise ValueError(f"Cannot combine an AnatomicalAnchor with {other.__class__}")

        if self.species != other.species:
            raise ValueError("Cannot combine an AnatomicalAnchor from different species.")
        else:
            species = self.species.union(other.species)

        regions = self.regions
        regions.update(other.regions)

        location = Location.union(self.location, other.location)

        return AnatomicalAnchor(species, location, regions)

    def __radd__(self, other) -> 'AnatomicalAnchor':
        # required to enable `sum`
        return self if other == 0 else self.__add__(other)



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handles reading and preparing gii files."""

from typing import Union, Dict

import numpy as np

from . import provider as _provider
from ...retrieval import requests
from ...commons import logger, merge_meshes
from ...locations import boundingbox as _boundingbox


class GiftiMesh(_provider.VolumeProvider, srctype="gii-mesh"):
    """
    One or more surface mesh fragments in Gifti format.
    """

    def __init__(self, url: Union[str, Dict[str, str]], volume=None):
        self._init_url = url
        self.volume = volume
        # TODO duplicated code to NgMesh
        if isinstance(url, str):  # single mesh
            self._loaders = {None: requests.HttpRequest(url)}
        elif isinstance(url, dict):   # named mesh fragments
            self._loaders = {lbl: requests.HttpRequest(u) for lbl, u in url.items()}
        else:
            raise NotImplementedError(f"Urls for {self.__class__.__name__} are expected to be of type str or dict.")

    @property
    def _url(self) -> Union[str, Dict[str, str]]:
        return self._init_url

    def get_boundingbox(self, clip=False, background=0.0, **fetch_kwargs) -> '_boundingbox.BoundingBox':
        """
        Bounding box calculation is not yet implemented for meshes.
        """
        raise NotImplementedError(
            f"Bounding box access to {self.__class__.__name__} objects not yet implemented."
        )

    @property
    def fragments(self):
        return [k for k in self._loaders if k is not None]

    def fetch(self, fragment: str = None, **kwargs):
        """
        Returns the mesh as a dictionary with two numpy arrays.

        Parameters
        ----------
        fragment: str, default: None
            A fragment name can be specified to choose from multiple fragments.

            Note
            ----
            If not specified, multiple fragments will be merged into one mesh.
            In such a case, the verts and faces arrays of different fragments
            are appended to one another.

        Returns
        -------
        dict
            - 'verts': An Nx3 array of vertex coordinates,
            - 'faces': an Mx3 array of face definitions using row indices of the vertex array
        """
        for arg in ["resolution_mm", "voi"]:
            if kwargs.get(arg):
                raise NotImplementedError(f"Parameter {arg} ignored by {self.__class__}.")

        fragments_included = []
        meshes = []
        for fragment_name, loader in self._loaders.items():
            if fragment and fragment.lower() not in fragment_name.lower():
                continue
            assert len(loader.data.darrays) > 1
            meshes.append({
                "verts": loader.data.darrays[0].data,
                "faces": loader.data.darrays[1].data
            })
            fragments_included.append(fragment_name)

        if len(fragments_included) > 1:
            logger.info(
                f"The mesh fragments [{', '.join(fragments_included)}] were merged by "
                "appending vertex information of fragments. "
                f"You could select one with the 'fragment' parameter in fetch()."
            )

        return merge_meshes(meshes)

    @property
    def variants(self):
        return list(self._loaders.keys())

    def fetch_iter(self):
        """
        Iterator returning all submeshes

        Returns
        -------
        dict
            - 'verts': An Nx3 array of vertex coordinates,
            - 'faces': an Mx3 array of face definitions using row indices of the vertex array
            - 'name': Name of the of the mesh variant
        """
        return (self.fetch(v) for v in self.variants)


class GiftiSurfaceLabeling(_provider.VolumeProvider, srctype="gii-label"):
    """
    A mesh labeling, specified by a gifti file.
    """

    def __init__(self, url: Union[str, dict]):
        self._init_url = url
        if isinstance(url, str):  # single mesh labelling
            self._loaders = {None: requests.HttpRequest(url)}
        elif isinstance(url, dict):   # labelling for multiple mesh fragments
            self._loaders = {lbl: requests.HttpRequest(u) for lbl, u in url.items()}
        else:
            raise NotImplementedError(f"Urls for {self.__class__.__name__} are expected to be of type str or dict.")

    def fetch(self, fragment: str = None, label: int = None, **kwargs):
        """Returns a 1D numpy array of label indices."""
        labels = []
        if fragment is None:
            matched_frags = list(self._loaders.keys())
        else:
            matched_frags = [frg for frg in self._loaders.keys() if fragment.lower() in frg.lower()]
            if len(matched_frags) != 1:
                raise ValueError(
                    f"Requested fragment '{fragment}' could not be matched uniquely "
                    f"to [{', '.join(self._loaders)}]"
                )
        for frag in matched_frags:
            assert len(self._loaders[frag].data.darrays) == 1
            if label is not None:
                labels.append((self._loaders[frag].data.darrays[0].data == label).astype('uint8'))
            else:
                labels.append(self._loaders[frag].data.darrays[0].data)

        return {"labels": np.hstack(labels)}

    def get_boundingbox(self, clip=False, background=0.0) -> '_boundingbox.BoundingBox':
        raise NotImplementedError(
            f"Bounding box access to {self.__class__.__name__} objects not yet implemented."
        )

    @property
    def _url(self) -> Union[str, Dict[str, str]]:
        return self._init_url



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Singular coordinate defined on a space, possibly with an uncertainty."""

from urllib.parse import quote
import re
import json
import numbers
import hashlib
from typing import Tuple, Union, Dict, TYPE_CHECKING

import numpy as np

from . import location, pointcloud
from ..commons import logger
from ..retrieval.requests import HttpRequest
from ..exceptions import SpaceWarpingFailedError, NoneCoordinateSuppliedError

if TYPE_CHECKING:
    from ..core.space import Space


class Point(location.Location):
    """A single 3D point in reference space."""

    @staticmethod
    def parse(spec, unit="mm") -> Tuple[float, float, float]:
        """Converts a 3D coordinate specification into a 3D tuple of floats.

        Parameters
        ----------
        spec: Any of str, tuple(float,float,float)
            For string specifications, comma separation with decimal points are expected.
        unit: str
            specification of the unit (only 'mm' supported so far)
        Returns
        -------
        tuple(float, float, float)
        """
        if unit != "mm":
            raise NotImplementedError(
                "Coordinate parsing from strings is only supported for mm specifications so far."
            )
        if isinstance(spec, str):
            pat = r"([-\d\.]*)" + unit
            digits = re.findall(pat, spec)
            if len(digits) == 3:
                return tuple(float(d) for d in digits)
        elif isinstance(spec, (tuple, list)) and len(spec) in [3, 4]:
            if any(v is None for v in spec):
                raise NoneCoordinateSuppliedError("Cannot parse cooridantes containing None values.")
            if len(spec) == 4:
                assert spec[3] == 1
            return tuple(float(v.item()) if isinstance(v, np.ndarray) else float(v) for v in spec[:3])
        elif isinstance(spec, np.ndarray) and spec.size == 3:
            if any(np.isnan(v) for v in spec):
                raise NoneCoordinateSuppliedError("Cannot parse cooridantes containing NaN values.")
            return tuple(float(v.item()) if isinstance(v, np.ndarray) else float(v) for v in spec[:3])
        elif isinstance(spec, Point):
            return spec.coordinate

        raise ValueError(
            f"Cannot decode the specification {spec} (type {type(spec)}) to create a point."
        )

    def __init__(
        self,
        coordinatespec,
        space=None,
        sigma_mm: float = 0.0,
        label: Union[int, float, tuple] = None
    ):
        """
        Construct a new 3D point set in the given reference space.

        Parameters
        ----------
        coordinatespec: 3-tuple of int/float, or string specification
            Coordinate in mm of the given space
        space: Space or str
            The reference space (id, object, or name)
        sigma_mm : float, optional
            Location uncertainty of the point
        label: optional
            Any object attached as an attribute to the point

            Note
            ----
                Interpreted as the isotropic standard deviation of location.
        """
        location.Location.__init__(self, space)
        self.coordinate = Point.parse(coordinatespec)
        self.sigma = sigma_mm
        self.label = label
        if isinstance(coordinatespec, Point):
            assert coordinatespec.sigma == sigma_mm
            assert coordinatespec.space == space
        self.label = label

    @property
    def homogeneous(self):
        """The homogeneous coordinate of this point as a 4-tuple,
        obtained by appending '1' to the original 3-tuple."""
        return np.atleast_2d(self.coordinate + (1,))

    def intersection(self, other: location.Location) -> "Point":
        if isinstance(other, Point):
            return self if self == other else None
        else:
            return self if other.intersection(self) else None

    def warp(self, space: Union[str, Dict, "Space"]):
        """
        Creates a new point by warping this point to another space
        TODO this needs to maintain the sigma parameter!
        """
        from ..core.space import Space

        spaceobj = space if isinstance(space, Space) else Space.get_instance(space)
        if spaceobj == self.space:
            return self
        if any(_ not in location.Location.SPACEWARP_IDS for _ in [self.space.id, spaceobj.id]):
            raise SpaceWarpingFailedError(
                f"Cannot convert coordinates between {self.space.id} and {spaceobj.id}"
            )
        url = "{server}/transform-point?source_space={src}&target_space={tgt}&x={x}&y={y}&z={z}".format(
            server=location.Location.SPACEWARP_SERVER,
            src=quote(location.Location.SPACEWARP_IDS[self.space.id]),
            tgt=quote(location.Location.SPACEWARP_IDS[spaceobj.id]),
            x=self.coordinate[0],
            y=self.coordinate[1],
            z=self.coordinate[2],
        )
        response = HttpRequest(url, lambda b: json.loads(b.decode())).get()
        if np.any(np.isnan(response['target_point'])):
            raise SpaceWarpingFailedError(f'Warping {str(self)} to {spaceobj.name} resulted in NaN')

        return self.__class__(
            coordinatespec=tuple(response["target_point"]),
            space=spaceobj.id,
            label=self.label
        )

    @property
    def volume(self):
        """ The volume of a point can be nonzero if it has a location uncertainty. """
        return self.sigma**3 * np.pi * 4. / 3.

    def __sub__(self, other):
        """Subtract the coordinates of two points to get
        a new point representing the offset vector. Alternatively,
        subtract an integer from the all coordinates of this point
        to create a new one.
        TODO this needs to maintain sigma
        """
        if isinstance(other, numbers.Number):
            return Point([c - other for c in self.coordinate], self.space)

        assert self.space == other.space
        return Point(
            [self.coordinate[i] - other.coordinate[i] for i in range(3)],
            self.space,
            label=self.label
        )

    def __lt__(self, other):
        o = other if self.space is None else other.warp(self.space)
        if o is None:
            return True  # 'other' was warped outside reference space bounds
        return all(self[i] < o[i] for i in range(3))

    def __gt__(self, other):
        assert other is not None
        o = other if self.space is None else other.warp(self.space)
        if o is None:
            return False  # 'other' was warped outside reference space bounds
        return all(self[i] > o[i] for i in range(3))

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other: 'Point'):
        if isinstance(other, pointcloud.PointCloud):
            return other == self  # implemented at pointcloud
        if not isinstance(other, Point):
            return False
        o = other if self.space is None else other.warp(self.space)
        if o is None:
            return False  # 'other' was warped outside reference space bounds
        return all(self[i] == o[i] for i in range(3)) and self.sigma == other.sigma

    def __le__(self, other):
        o = other if self.space is None else other.warp(self.space)
        if o is None:
            return True  # 'other' was warped outside reference space bounds
        return all(self[i] <= o[i] for i in range(3))

    def __ge__(self, other):
        assert other is not None
        o = other if self.space is None else other.warp(self.space)
        if o is None:
            return False  # 'other' was warped outside reference space bounds
        return all(self[i] >= o[i] for i in range(3))

    def __add__(self, other):
        """Add the coordinates of two points to get
        a new point representing."""
        if isinstance(other, numbers.Number):
            return Point([c + other for c in self.coordinate], self.space)
        if isinstance(other, Point):
            assert self.space == other.space
        return Point(
            [self.coordinate[i] + other.coordinate[i] for i in range(3)],
            self.space,
            sigma_mm=self.sigma + other.sigma,
            label=(self.label, other.label)
        )

    def __truediv__(self, number: float):
        """Return a new point with divided
        coordinates in the same space."""
        return Point(
            np.array(self.coordinate) / number,
            self.space,
            sigma_mm=self.sigma / number,
            label=self.label
        )

    def __mul__(self, number: float):
        """Return a new point with multiplied
        coordinates in the same space."""
        return Point(
            np.array(self.coordinate) * number,
            self.space,
            sigma_mm=self.sigma * number,
            label=self.label
        )

    def transform(self, affine: np.ndarray, space=None):
        """Returns a new Point obtained by transforming the
        coordinate of this one with the given affine matrix.
        TODO this needs to maintain sigma

        Parameters
        ----------
        affine: numpy 4x4 ndarray
            affine matrix
        space: str, Space, or None
            Target reference space which is reached after applying the transform

            Note
            ----
            The consistency of this cannot be checked and is up to the user.
        """
        x, y, z, h = np.dot(affine, self.homogeneous.T)
        if h != 1:
            logger.warning(f"Homogeneous coordinate is not one: {h}")
        return self.__class__(
            (x / h, y / h, z / h),
            space,
            sigma_mm=self.sigma,
            label=self.label
        )

    def get_enclosing_cube(self, width_mm):
        """
        Create a bounding box centered around this point with the given width.
        TODO this should respect sigma (in addition or instead of the offset)
        """
        offset = width_mm / 2
        from .boundingbox import BoundingBox
        return BoundingBox(
            point1=self - offset,
            point2=self + offset,
            space=self.space,
        )

    def __len__(self):
        return 1

    def __iter__(self):
        """Return an iterator over the location,
        so the Point can be easily cast to list or tuple."""
        return iter(self.coordinate)

    def __setitem__(self, index, value):
        """Write access to the coefficients of this point."""
        assert 0 <= index < 3
        values = list(self.coordinate)
        values[index] = value
        self.coordinate = tuple(values)

    def __getitem__(self, index):
        """Index access to the coefficients of this point."""
        assert 0 <= index < 3
        return self.coordinate[index]

    def bigbrain_section(self):
        """
        Estimate the histological section number of BigBrain
        which corresponds to this point. If the point is given
        in another space, a warping to BigBrain space will be tried.
        """
        if self.space.id == location.Location.BIGBRAIN_ID:
            coronal_position = self[1]
        else:
            try:
                bigbrain_point = self.warp("bigbrain")
                coronal_position = bigbrain_point[1]
            except Exception:
                raise RuntimeError(
                    "BigBrain section numbers can only be determined "
                    "for points in BigBrain space, but the given point "
                    f"is given in '{self.space.name}' and could not "
                    "be converted."
                )
        return int((coronal_position + 70.0) / 0.02 + 1.5)

    @property
    def id(self) -> str:
        return hashlib.md5(
            f"{self.space.id}{','.join(str(val) for val in self)}".encode("utf-8")
        ).hexdigest()

    def __repr__(self):
        spacespec = f"'{self.space.id}'" if self.space else None
        return f"<Point({self.coordinate}, space={spacespec}, sigma_mm={self.sigma})>"



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from math import atan2

import numpy as np
from nilearn import image

from . import point, pointcloud, boundingbox
from ..volumes import volume
from ..commons import translation_matrix, y_rotation_matrix


class AxisAlignedPatch(pointcloud.PointCloud):

    def __init__(self, corners: pointcloud.PointCloud):
        """Construct a patch in physical coordinates.
        As of now, only patches aligned in the y plane of the physical space
        are supported."""
        # TODO: need to ensure that the points are planar, if more than 3
        assert len(corners) == 4
        assert len(np.unique(corners.coordinates[:, 1])) == 1
        pointcloud.PointCloud.__init__(
            self,
            coordinates=corners.coordinates,
            space=corners.space,
            sigma_mm=corners.sigma_mm,
            labels=corners.labels,
        )

    def __str__(self):
        return f"Patch with boundingbox {self.boundingbox}"

    def flip(self):
        """Returns a flipped version of the patch."""
        new_corners = self.coordinates.copy()[[2, 3, 0, 1]]
        return AxisAlignedPatch(pointcloud.PointCloud(new_corners, self.space))

    def extract_volume(
        self,
        image_volume: volume.Volume,
        resolution_mm: float,
    ):
        """
        fetches image data in a planar patch.
        TODO The current implementation only covers patches which are strictly
        define in the y plane. A future implementation should accept arbitrary
        oriented patches.accept arbitrary oriented patches.
        """
        assert image_volume.space == self.space

        # Extend the 2D patch into a 3D structure
        # this is only valid if the patch plane lies within the image canvas.
        canvas = image_volume.get_boundingbox()
        assert canvas.minpoint[1] <= self.coordinates[0, 1]
        assert canvas.maxpoint[1] >= self.coordinates[0, 1]
        XYZ = self.coordinates
        voi = boundingbox.BoundingBox(
            XYZ.min(0)[:3], XYZ.max(0)[:3], space=image_volume.space
        )
        # enforce the patch to have the same y dimensions
        voi.minpoint[1] = canvas.minpoint[1]
        voi.maxpoint[1] = canvas.maxpoint[1]
        patch = image_volume.fetch(voi=voi, resolution_mm=resolution_mm)
        assert patch is not None

        # patch rotation defined in physical space
        vx, vy, vz = XYZ[1] - XYZ[0]
        alpha = -atan2(-vz, -vx)
        cx, cy, cz = XYZ.mean(0)
        rot_phys = np.linalg.multi_dot(
            [
                translation_matrix(cx, cy, cz),
                y_rotation_matrix(alpha),
                translation_matrix(-cx, -cy, -cz),
            ]
        )

        # rotate the patch in physical space
        affine_rot = np.dot(rot_phys, patch.affine)

        # crop in the rotated space
        pixels = np.dot(np.linalg.inv(affine_rot), self.homogeneous.T).astype("int").T
        # keep a pixel distance to avoid black border pixels
        xmin, ymin, zmin = pixels.min(0)[:3] + 1
        xmax, ymax, zmax = pixels.max(0)[:3] - 1
        h, w = xmax - xmin, zmax - zmin
        affine = np.dot(affine_rot, translation_matrix(xmin, 0, zmin))
        return volume.from_nifti(
            image.resample_img(
                patch,
                target_affine=affine,
                target_shape=[h, 1, w],
                force_resample=True
            ),
            space=image_volume.space,
            name=f"Rotated patch with corner points {self.coordinates} sampled from {image_volume.name}",
        )


class Plane:
    """
    A 3D plane in reference space.
    TODO This shall eventually be derived from siibra.Location
    """

    def __init__(self, point1: point.Point, point2: point.Point, point3: point.Point):
        """
        Create a 3D plane from 3 points.
        The plane's reference space is defined by the first point.
        """
        self.space = point1.space
        # normal is the cross product of two arbitrary in-plane vectors
        n = np.cross(
            (point2.warp(self.space) - point1).coordinate,
            (point3.warp(self.space) - point1).coordinate,
        )
        self._n = n / np.linalg.norm(n)
        # distance from origin is the projection of any in-plane point onto the normal
        self._d = np.dot(point1.coordinate, self._n)

    @property
    def normal(self):
        return self._n

    @property
    def distance_from_origin(self):
        return self._d

    def sidedness(self, points: np.ndarray):
        """
        Computes labels for a set of 3D coordinates classifying them
        by the halfspaces spanned by this plane.
        """
        assert points.shape[1] == 3
        return (np.dot(points, self.normal) >= self.distance_from_origin).astype("int")

    def intersect_line_segments(self, startpoints: np.ndarray, endpoints: np.ndarray):
        """
        Intersects a set of straight line segments with the plane.
        Returns the set of intersection points.
        The line segments are given by two Nx3 arrays of their start- and endpoints.
        The result is an Nx3 list of intersection coordinates.
        """
        directions = endpoints - startpoints
        lengths = np.linalg.norm(directions, axis=1)
        directions = directions / lengths[:, None]
        lambdas = (self._d - np.dot(startpoints, self._n)) / np.dot(directions, self._n)
        assert all(lambdas >= 0)
        result = startpoints + lambdas[:, None] * directions
        non_intersecting = lambdas > lengths
        num_failed = np.count_nonzero(non_intersecting)
        result[non_intersecting, :] = np.nan
        if num_failed > 0:
            print(
                "WARNING: line segment intersection includes NaN rows "
                f"for {num_failed} non-intersecting segments."
            )
        return result

    def intersect_mesh(self, mesh: dict) -> List[pointcloud.Contour]:
        """
        Intersects a 3D surface mesh with the plane.
        Returns a set of split 2D contours, represented by ordered coordinate lists.
        The output contour coordinates are intersection points of mesh edges going through the plane,
        and only combined into a contour if arising from neighboring edges in the mesh.
        The mesh is passed as a dictionary with an Nx3 array "verts" of vertex coordinates,
        and an Mx3 array "faces" of face definitions.
        Each row in the face array corresponds to the three indices of vertices making up the
        triangle.
        The result is a list of contour segments, each represented as a PointCloud
        holding the ordered list of contour points.
        The point labels in each "contour" PointCloud hold the index of the face in the
        mesh which made up each contour point.
        """

        # select faces whose vertices are in different halfspaces relative to the y plane
        vertex_in_halfspace = self.sidedness(mesh["verts"])
        face_vertex_in_halfspace = vertex_in_halfspace[mesh["faces"]]
        face_indices = np.where(
            face_vertex_in_halfspace.min(1) != face_vertex_in_halfspace.max(1)
        )[0]
        faces = mesh["faces"][face_indices]

        # for each of N selected faces, indicate whether we cross the plane
        # as we go from vertex 2->0, 0->1, 1->2, respectively.
        # This gives us an Nx3 array, where forward crossings are identified by 1,
        # and backward crossings by -1.
        # Each column of the crossings is linked to two columns of the faces array.
        crossings = np.diff(
            face_vertex_in_halfspace[face_indices][:, [2, 0, 1, 2]], axis=1
        )
        face_columns = np.array([[2, 0], [0, 1], [1, 2]])

        # We assume that there is exactly one forward and one inverse crossing
        # per selected face. Test this assumption.
        # NOTE This will fail if an edge is exactly in-plane
        assert all(all((crossings == v).sum(1) == 1) for v in [-1, 0, 1])

        # Compute the actual intersection points for forward and backward crossing edges.
        fwd_columns = np.where(crossings == 1)[1]
        bwd_columns = np.where(crossings == -1)[1]
        fwd_vertices = np.array(
            [
                faces[np.arange(len(faces)), face_columns[fwd_columns][:, 0]],
                faces[np.arange(len(faces)), face_columns[fwd_columns][:, 1]],
            ]
        ).T
        bwd_vertices = np.array(
            [
                faces[np.arange(len(faces)), face_columns[bwd_columns][:, 1]],
                faces[np.arange(len(faces)), face_columns[bwd_columns][:, 0]],
            ]
        ).T
        fwd_intersections = self.intersect_line_segments(
            mesh["verts"][fwd_vertices[:, 0]], mesh["verts"][fwd_vertices[:, 1]]
        )
        bwd_intersections = self.intersect_line_segments(
            mesh["verts"][bwd_vertices[:, 0]], mesh["verts"][bwd_vertices[:, 1]]
        )

        # By construction, the fwd and backward intersections
        # should include the exact same set of points. Verify this now.
        sortrows = lambda A: A[np.lexsort(A.T[::-1]), :]
        err = (sortrows(fwd_intersections) - sortrows(bwd_intersections)).sum()
        assert err == 0, f"intersection inconsistency: {err}"

        # Due to the above property, we can construct closed contours in the
        # intersection plane by following the interleaved fwd/bwd roles of intersection
        # points.
        face_indices = list(range(fwd_intersections.shape[0]))
        result = []
        points = []
        labels = []
        face_id = 0  # index of the mesh face to consider
        while len(face_indices) > 0:

            # continue the contour with the next forward edge intersection
            p = fwd_intersections[face_id]
            points.append(p)
            # Remember the ids of the face and start-/end vertices for the point
            labels.append((face_id, fwd_vertices[face_id, 0], fwd_vertices[face_id, 1]))
            face_indices.remove(face_id)
            neighbours = np.where(np.all(np.isin(bwd_intersections, p), axis=1))[0]
            assert len(neighbours) > 0
            face_id = neighbours[0]
            if face_id in face_indices:
                # more points available in the contour
                continue

            # finish the current contour.
            result.append(
                pointcloud.Contour(np.array(points), labels=labels, space=self.space)
            )
            if len(face_indices) > 0:
                # prepare to process another contour segment
                face_id = face_indices[0]
                points = []
                labels = []

        return result

    def project_points(self, points: pointcloud.PointCloud):
        """projects the given points onto the plane."""
        assert self.space == points.space
        XYZ = points.coordinates
        N = XYZ.shape[0]
        dists = np.dot(self._n, XYZ.T) - self._d
        return pointcloud.PointCloud(
            XYZ - np.tile(self._n, (N, 1)) * dists[:, np.newaxis], space=self.space
        )

    def get_enclosing_patch(self, points: pointcloud.PointCloud, margin=[0.5, 0.5]):
        """
        Computes the enclosing patch in the given plane
        which contains the projections of the given points.
        The orientation of the patch follows the principal axis.
        """
        projections = self.project_points(points)

        # compute PCA of point projections to obtain two orthogonal
        # in-plane spanning vectors
        XYZ = np.copy(projections.coordinates)
        mu = XYZ.mean(0)
        XYZ -= mu
        cov = np.dot(XYZ.T, XYZ)
        eigvals_, eigvecs_ = np.linalg.eigh(cov)
        eigvecs = eigvecs_[:, ::-1].T
        v1, v2 = [-eigvecs[_] for _ in np.argsort(eigvals_)[:2]]

        # get extremal points along first spanning vector
        order = np.argsort(np.dot(projections.coordinates, v1))
        p0 = projections[order[0]].homogeneous[0, :3]
        p1 = projections[order[-1]].homogeneous[0, :3]

        # get extremal points along second spanning vector
        order = np.argsort(np.dot(projections.coordinates, v2))
        p2 = projections[order[0]].homogeneous[0, :3]
        p3 = projections[order[-1]].homogeneous[0, :3]

        m0, m1 = margin
        w = np.linalg.norm(p3 - p2)
        corners = pointcloud.PointCloud(
            [
                p1 + (w / 2 + m1) * v2 + m0 * v1,
                p0 + (w / 2 + m1) * v2 - m0 * v1,
                p0 - (w / 2 + m1) * v2 - m0 * v1,
                p1 - (w / 2 + m1) * v2 + m0 * v1,
            ],
            space=self.space,
        )
        err = (self.project_points(corners).coordinates - corners.coordinates).sum()
        if err > 1e-5:
            print(
                f"WARNING: patch coordinates were not exactly in-plane (error={err})."
            )

        try:
            patch = AxisAlignedPatch(self.project_points(corners))
        except AssertionError:
            patch = None
        return patch

    @classmethod
    def from_image(cls, image: volume.Volume):
        """
        Derive an image plane by assuming the volume to be a 2D image.
        The smallest dimension in voxel space is considered flat.
        The plane is defined in the physical space of the volume.
        """
        assert isinstance(image, volume.Volume)
        im_lowres = image.fetch(resolution_mm=1)
        plane_dims = np.where(np.argsort(im_lowres.shape) < 2)[0]
        voxels = pointcloud.PointCloud(
            np.vstack(([0, 0, 0], np.identity(3)[plane_dims])), space=None
        )
        points = voxels.transform(im_lowres.affine, space=image.space)
        return cls(points[0], points[1], points[2])



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handles spatial concepts and spatial operation like warping between spaces."""
from typing import Iterable
from functools import reduce

from .location import Location
from .point import Point
from .pointcloud import PointCloud, Contour, from_points
from .experimental import AxisAlignedPatch, Plane
from .boundingbox import BoundingBox


def reassign_union(*args: Iterable["Location"]) -> "Location":
    return reduce(pairwise_union, args)


def pairwise_union(loc0: "Location", loc1: "Location") -> "Location":
    """
    Add two locations of same or different type to find their union as a
    Location object.

    Note
    ----
    `loc1` will be warped to `loc0` they are not in the same space.

    Parameters
    ----------
    loc0 : Location
    loc1 : Location

    Returns
    -------
    Location
        - Point U Point = PointCloud
        - Point U PointCloud = PointCloud
        - PointCloud U PointCloud = PointCloud
        - BoundingBox U BoundingBox = BoundingBox
        - BoundingBox U PointCloud = BoundingBox
        - BoundingBox U Point = BoundingBox
        - WholeBrain U Location = NotImplementedError
        (all operations are commutative)
    """
    if loc0 is None or loc1 is None:
        return loc0 or loc1

    # All location types should be unionable among each other and this should
    # be implemented here to avoid code repetition. Volumes are the only type of
    # location that has its own union method since it is not a part of locations
    # module and to avoid importing Volume here.
    if not all(
        isinstance(loc, (Point, PointCloud, BoundingBox)) for loc in [loc0, loc1]
    ):
        try:
            return loc1.union(loc0)
        except Exception:
            raise NotImplementedError(
                f"There are no union method for {(loc0.__class__.__name__, loc1.__class__.__name__)}"
            )

    # convert Points to PointClouds
    loc0, loc1 = [
        from_points([loc]) if isinstance(loc, Point) else loc for loc in [loc0, loc1]
    ]

    # adopt the space of the first location
    loc1_w = loc1.warp(loc0.space)

    if isinstance(loc0, PointCloud):
        if isinstance(loc1_w, PointCloud):
            points = list(dict.fromkeys([*loc0, *loc1_w]))
            return from_points(points)
        if isinstance(loc1_w, BoundingBox):
            return reassign_union(loc0.boundingbox, loc1_w)

    if isinstance(loc0, BoundingBox) and isinstance(loc1_w, BoundingBox):
        coordinates = [loc0.minpoint, loc0.maxpoint, loc1_w.minpoint, loc1_w.maxpoint]
        return BoundingBox(
            point1=[min(p[i] for p in coordinates) for i in range(3)],
            point2=[max(p[i] for p in coordinates) for i in range(3)],
            space=loc0.space,
            sigma_mm=[loc0.minpoint.sigma, loc0.maxpoint.sigma],
        )

    return reassign_union(loc1_w, loc0)


Location.union = reassign_union



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Concepts that have primarily spatial meaning."""

from __future__ import annotations

from typing import Union, Dict
from abc import abstractmethod

import numpy as np

from ..core.structure import BrainStructure
from ..core import space as _space


class Location(BrainStructure):
    """
    Abstract base class for locations in a given reference space.
    """

    # backend for transforming coordinates between spaces
    # for detail, see https://github.com/HumanBrainProject/hbp-spatial-backend
    SPACEWARP_SERVER = "https://siibra-spatial-backend.apps.ebrains.eu/v1"

    # lookup of space identifiers to be used by SPACEWARP_SERVER
    SPACEWARP_IDS = {
        "minds/core/referencespace/v1.0.0/dafcffc5-4826-4bf1-8ff6-46b8a31ff8e2": "MNI 152 ICBM 2009c Nonlinear Asymmetric",
        "minds/core/referencespace/v1.0.0/7f39f7be-445b-47c0-9791-e971c0b6d992": "MNI Colin 27",
        "minds/core/referencespace/v1.0.0/a1655b99-82f1-420f-a3c2-fe80fd4c8588": "Big Brain (Histology)",
    }

    # The id of BigBrain reference space
    BIGBRAIN_ID = "minds/core/referencespace/v1.0.0/a1655b99-82f1-420f-a3c2-fe80fd4c8588"
    _MASK_MEMO = {}  # cache region masks for Location._assign_region()
    _ASSIGNMENT_CACHE = {}  # caches assignment results, see Region.assign()

    def __init__(self, spacespec: Union[str, Dict[str, str], _space.Space]):
        self._space_spec = spacespec
        self._space_cached = None

    @property
    def space(self) -> _space.Space:
        if self._space_cached is None:
            if self._space_spec is None:
                return None
            elif isinstance(self._space_spec, _space.Space):
                self._space_cached = self._space_spec
            elif isinstance(self._space_spec, dict):
                spec = self._space_spec.get("@id") or self._space_spec.get("name")
                self._space_cached = _space.Space.get_instance(spec)
            elif isinstance(self._space_spec, str):
                self._space_cached = _space.Space.get_instance(self._space_spec)
            else:
                raise ValueError(f"Invalid space spec type: '{type(self._space_spec)}'")
        return self._space_cached

    @abstractmethod
    def warp(self, space: Union[str, "_space.Space"]):
        """Generates a new location by warping the
        current one into another reference space."""
        pass

    @abstractmethod
    def transform(self, affine: np.ndarray, space: Union[str, "_space.Space", None] = None):
        """Returns a new location obtained by transforming the
        reference coordinates of this one with the given affine matrix.

        Parameters
        ----------
        affine : numpy 4x4 ndarray
            affine matrix
        space : reference space (id, name, or Space)
            Target reference space which is reached after
            applying the transform. Note that the consistency
            of this cannot be checked and is up to the user.
        """
        pass

    @property
    def species(self):
        return None if self.space is None else self.space.species

    def __str__(self):
        space_str = "" if self.space is None else f" in {self.space.name}"
        coord_str = "" if len(self) == 0 else f" [{','.join(str(pt) for pt in iter(self))}]"
        return f"{self.__class__.__name__}{space_str}{coord_str}"

    def __repr__(self):
        spacespec = f"'{self.space.id}'" if self.space else None
        return f"<{self.__class__.__name__}({[point.__repr__() for point in self]}), space={spacespec}>"

    def __hash__(self) -> int:
        return hash(self.__repr__())

    @abstractmethod
    def __eq__(self):
        """Required to provide comparison and making the object hashable"""
        raise NotImplementedError

    @staticmethod
    def union(loc0: 'Location', loc1: 'Location') -> 'Location':
        """
        Reassigned at the locations module level for static typing and to avoid
        circular imports. See siibra.locations.__init__.reassign_union()
        """
        raise NotImplementedError(
            "This method is designed to be reassigned at the module level"
        )



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, TYPE_CHECKING
from urllib.parse import quote_plus
from numpy import int32
import numpy as np
import re
from dataclasses import dataclass
import math

from .util import encode_number, separator, cipher, neg, decode_number, post_process

if TYPE_CHECKING:
    from siibra.core.atlas import Atlas
    from siibra.core.space import Space
    from siibra.locations import BoundingBox, Point
    from siibra.core.parcellation import Parcellation
    from siibra.core.region import Region
    from siibra.features.feature import Feature


class DecodeNavigationException(Exception):
    pass


min_int32 = -2_147_483_648
max_int32 = 2_147_483_647


default_root_url = "https://atlases.ebrains.eu/viewer/"


def sanitize_id(id: str):
    return id.replace("/", ":")


def get_perspective_zoom(
    atlas: "Atlas", space: "Space", parc: "Parcellation", region: Optional["Region"]
):
    import siibra

    if atlas is siibra.atlases["rat"] or atlas is siibra.atlases["mouse"]:
        return 200000
    return 2000000


def get_zoom(
    atlas: "Atlas", space: "Space", parc: "Parcellation", region: Optional["Region"]
):
    import siibra

    if atlas is siibra.atlases["rat"] or atlas is siibra.atlases["mouse"]:
        return 35000
    return 350000


supported_prefix = ("nifti://", "swc://", "precomputed://", "deepzoom://")


def append_query_params(url: str, *args, query_params={}, **kwargs):
    query_str = "&".join(
        [f"{key}={quote_plus(value)}" for key, value in query_params.items()]
    )
    if len(query_str) > 0:
        query_str = "?" + query_str
    return url + query_str


@post_process(append_query_params)
def encode_url(
    atlas: "Atlas",
    space: "Space",
    parc: "Parcellation",
    region: Optional["Region"] = None,
    *,
    root_url=default_root_url,
    external_url: str = None,
    location: "Point" = None,
    feature: "Feature" = None,
    ignore_warning=False,
    query_params={},
):
    from siibra.locations import Point
    overlay_url = None
    encoded_position = None
    if location:
        assert isinstance(location, Point), "currently, location only supports Point"
        encoded_position = ".".join([encode_number(int(p * 1e6)) for p in location])
    if external_url:
        assert any(
            [external_url.startswith(prefix) for prefix in supported_prefix]
        ), f"url needs to start with {(' , '.join(supported_prefix))}"
        overlay_url = "/x-overlay-layer:{url}".format(
            url=external_url.replace("/", "%2F")
        )

    zoom = get_zoom(atlas, space, parc, region)
    pzoom = get_perspective_zoom(atlas, space, parc, region)

    zoom_kwargs = {
        "encoded_pzoom": encode_number(pzoom, False),
        "encoded_zoom": encode_number(zoom, False),
    }
    nav_string = "/@:0.0.0.-W000.._eCwg.2-FUe3._-s_W.2_evlu..{encoded_pzoom}..{encoded_nav}..{encoded_zoom}"

    return_url = (
        "{root_url}#/a:{atlas_id}/t:{template_id}/p:{parc_id}{overlay_url}".format(
            root_url=root_url,
            atlas_id=sanitize_id(atlas.id),
            template_id=sanitize_id(space.id),
            parc_id=sanitize_id(parc.id),
            overlay_url=overlay_url if overlay_url else "",
        )
    )

    if feature is not None:
        return_url = return_url + f"/f:{sanitize_id(feature.id)}"

    if region is None:
        return return_url + nav_string.format(encoded_nav=encoded_position or "0.0.0", **zoom_kwargs)

    return_url = f"{return_url}/rn:{get_hash(region.name)}"

    try:
        result_props = region.spatial_props(space, maptype="labelled")
        if len(result_props) == 0:
            return return_url + nav_string.format(encoded_nav=encoded_position or "0.0.0", **zoom_kwargs)
    except Exception as e:
        print(f"Cannot get_spatial_props {str(e)}")
        if not ignore_warning:
            raise e
        return return_url + nav_string.format(encoded_nav=encoded_position or "0.0.0", **zoom_kwargs)

    centroid = result_props[0].centroid

    encoded_centroid = separator.join(
        [encode_number(math.floor(val * 1e6)) for val in centroid]
    )
    return_url = return_url + nav_string.format(
        encoded_nav=encoded_position or encoded_centroid, **zoom_kwargs
    )
    return return_url


@dataclass
class DecodedUrl:
    bounding_box: "BoundingBox"


def decode_url(url: str, vp_length=1000):
    import siibra
    from siibra.locations import Point, BoundingBox

    try:
        space_match = re.search(r"/t:(?P<space_id>[^/]+)", url)
        space_id = space_match.group("space_id")
        space_id = space_id.replace(":", "/")
        space = siibra.spaces[space_id]
    except Exception as e:
        raise DecodeNavigationException from e

    nav_match = re.search(r"/@:(?P<navigation_str>[^/]+)/?", url)
    navigation_str = nav_match.group("navigation_str")
    for char in navigation_str:
        assert char in cipher or char in [
            neg,
            separator,
        ], f"char {char} not in cipher, nor separator/neg"

    try:
        ori_enc, pers_ori_enc, pers_zoom_enc, pos_enc, zoomm_enc = navigation_str.split(
            f"{separator}{separator}"
        )
    except Exception as e:
        raise DecodeNavigationException from e

    try:
        x_enc, y_enc, z_enc = pos_enc.split(separator)
        pos = [decode_number(val) for val in [x_enc, y_enc, z_enc]]
        zoom = decode_number(zoomm_enc)

        # zoom = nm/pixel
        pt1 = [(coord - (zoom * vp_length / 2)) / 1e6 for coord in pos]
        pt1 = Point(pt1, space)

        pt2 = [(coord + (zoom * vp_length / 2)) / 1e6 for coord in pos]
        pt2 = Point(pt2, space)

    except Exception as e:
        raise DecodeNavigationException from e

    bbx = BoundingBox(pt1, pt2, space)
    return DecodedUrl(bounding_box=bbx)


def get_hash(full_string: str):
    return_val = 0
    with np.errstate(over="ignore"):
        for char in full_string:
            # overflowing is expected and in fact the whole reason why convert number to int32

            # in windows, int32((0 - min_int32) << 5), rather than overflow to wrapper around, raises OverflowError
            shifted_5 = int32(
                (return_val - min_int32) if return_val > max_int32 else return_val << 5
            )

            return_val = int32(shifted_5 - return_val + ord(char))
            return_val = return_val & return_val
    hex_val = hex(return_val)
    return hex_val[3:]



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Query Allen Human Brain Atlas microarray data in specified volume."""

from typing import List
from xml.etree import ElementTree
import json

import numpy as np

from . import query as _query
from ..core import structure
from ..core.region import Region
from ..features import anchor as _anchor
from ..features.tabular.gene_expression import GeneExpressions
from ..commons import logger, Species
from ..locations import pointcloud
from ..retrieval import HttpRequest
from ..vocabularies import GENE_NAMES


BASE_URL = "http://api.brain-map.org/api/v2/data"

LOCATION_PRECISION_MM = 2.  # the assumed spatial precision of the probe locations in MNI space


def is_allen_api_microarray_service_available():
    import requests

    # see https://community.brain-map.org/t/human-brain-atlas-api/2876
    microarray_test_url = "http://api.brain-map.org/api/v2/data/query.json?criteria= service::human_microarray_expression[probes$eq1023146,1023147][donors$eq15496][structures$eq9148]"
    try:
        response = requests.get(microarray_test_url).json()
    except requests.RequestException:
        return False
    return response["success"]


def parse_gene(spec):
    if isinstance(spec, str):
        return [GENE_NAMES.get(spec)]
    elif isinstance(spec, dict):
        assert all(k in spec for k in ['symbol', 'description'])
        assert spec['symbol'] in GENE_NAMES
        return [spec]
    elif isinstance(spec, list):
        return [g for s in spec for g in parse_gene(s)]
    else:
        logger.error("Invalid specification of gene:", spec)
        return []


class InvalidAllenAPIResponseException(Exception):
    pass


class AllenBrainAtlasQuery(_query.LiveQuery, args=['gene'], FeatureType=GeneExpressions):
    """
    Interface to Allen Human Brain Atlas microarray data.

    This class connects to the web API of the Allen Brain Atlas:
    © 2015 Allen Institute for Brain Science. Allen Brain Atlas API.
    Available from: brain-map.org/api/index.html
    Any use of the data needs to be in accordance with their terms of use, see
    https://alleninstitute.org/legal/terms-use/

    - We have samples from 6 different human donors.
    - Each donor corresponds to exactly 1 specimen (tissue used for study)
    - Each sample was subject to multiple (in fact 4) different probes.
    - The probe data structures contain the list of gene expression of a
      particular gene measured in each sample. Therefore the length of the gene
      expression list in a probe corresponds to the number of samples taken in
      the corresponding donor for the given gene.
    """

    _FEATURETYPE = GeneExpressions

    _notification_shown = False

    _QUERY = {
        "probe": BASE_URL
        + "/query.xml?criteria=model::Probe,rma::criteria,[probe_type$eq'DNA'],products[abbreviation$eq'HumanMA'],gene[acronym$eq'{gene}'],rma::options[only$eq'probes.id']",
        "multiple_gene_probe": BASE_URL
        + "/query.xml?criteria=model::Probe,rma::criteria,[probe_type$eq'DNA'],products[abbreviation$eq'HumanMA'],gene[acronym$in{genes}],rma::options[only$eq'probes.id']&start_row={start_row}&num_rows={num_rows}",
        "specimen": BASE_URL
        + "/Specimen/query.json?criteria=[name$eq'{specimen_id}']&include=alignment3d",
        "microarray": BASE_URL
        + "/query.json?criteria=service::human_microarray_expression[probes$in{probe_ids}][donors$eq{donor_id}]",
        "gene": BASE_URL
        + "/Gene/query.json?criteria=products[abbreviation$eq'HumanMA']&num_rows=all",
        "factors": BASE_URL
        + "/query.json?criteria=model::Donor,rma::criteria,products[id$eq2],rma::include,age,rma::options[only$eq%27donors.id,dono  rs.name,donors.race_only,donors.sex%27]&start_row={start_row}&num_rows={num_rows}",
    }

    # there is a 1:1 mapping between donors and specimen for the 6 adult human brains
    _DONOR_IDS = ["15496", "14380", "15697", "9861", "12876", "10021"]
    _SPECIMEN_IDS = [
        "H0351.1015",
        "H0351.1012",
        "H0351.1016",
        "H0351.2001",
        "H0351.1009",
        "H0351.2002",
    ]

    _specimen = None
    factors = None
    species = Species.decode('homo sapiens')

    def __init__(self, **kwargs):
        """
        Each instance of this live query retrieves the probe IDs
        containing measurements for any gene in the given set
        of candidate genes.
        Each probe has expression levels and z-scores for a set of
        N samples.
        Each sample is linked to a donor, brain structure, and
        ICBM coordinate.
        When querying with a brain structure, the ICBM coordinates
        will be tested against the region mask in ICBM space
        to produce a table of outputs.
        """
        _query.LiveQuery.__init__(self, **kwargs)
        gene = kwargs.get('gene')
        if gene is None:
            self.genes = []
        else:
            self.genes = parse_gene(gene)

    def query(self, concept: structure.BrainStructure) -> List[GeneExpressions]:
        if not is_allen_api_microarray_service_available():
            raise InvalidAllenAPIResponseException(
                'The service "web API of the Allen Brain Atlas for the human microarray expression" '
                'is not available at the moment, therefore siibra is not able to fetch '
                'gene expression features. This is a known issue which we are investigating: '
                'https://github.com/FZJ-INM1-BDA/siibra-python/issues/636.'
            )
        if len(self.genes) == 0:
            logger.error("Cannot query for gene expressions without a specified gene.")
            return []

        # Match the microarray probes to the query mask.
        # Record matched instances and their locations.
        all_measurements = list(self)
        all_mes_points = pointcloud.PointCloud(
            [measurement['mni_xyz'] for measurement in all_measurements],
            space='mni152',
            sigma_mm=LOCATION_PRECISION_MM
        )
        intersecting_points = concept.intersection(all_mes_points)
        if intersecting_points is None:
            logger.info(f"No probes found that lie within {concept}")
            return []
        measurements = [all_measurements[index] for index in intersecting_points.labels]

        # Build the anatomical anchor and assignment to the query concept.
        # It will be attached to the returned feature, with the set of matched
        # MNI coordinates as anchor's location.
        anchor = _anchor.AnatomicalAnchor(
            location=intersecting_points,
            species=self.species
        )
        explanation = f"MNI coordinates of tissue samples were filtered using {concept}"
        anchor._assignments[concept] = [_anchor.AnatomicalAssignment(
            query_structure=concept,
            assigned_structure=concept,
            qualification=_anchor.Qualification.CONTAINED,
            explanation=explanation
        )]
        anchor._last_matched_concept = concept
        if isinstance(concept, Region):
            anchor._regionspec = concept.name

        return [GeneExpressions(
            anchor=anchor,
            genes=[m['gene'] for m in measurements],
            levels=[m['expression_level'] for m in measurements],
            z_scores=[m['z_score'] for m in measurements],
            additional_columns={
                "race": [m['race'] for m in measurements],
                "gender": [m['gender'] for m in measurements],
                "age": [m['age'] for m in measurements],
                "mni_xyz": [tuple(m['mni_xyz']) for m in measurements],
                "sample": [m['sample_index'] for m in measurements],
                "probe_id": [m['probe_id'] for m in measurements],
                "donor_name": [m['donor_name'] for m in measurements],
            }
        )]

    def __iter__(self):

        if self.genes is None:
            logger.warning(
                f"No gene name provided to {self.__class__.__name__}, so no gene expressions will be retrieved. "
                'Use the "gene=<name>" option in the feature query to specify one.'
            )
            return

        if not self.__class__._notification_shown:
            print(GeneExpressions.ALLEN_ATLAS_NOTIFICATION)
            self.__class__._notification_shown = True

        probe_ids = self._retrieve_probe_ids(self.genes)

        # get specimen information
        if AllenBrainAtlasQuery._specimen is None:
            AllenBrainAtlasQuery._specimen = {
                spcid: AllenBrainAtlasQuery._retrieve_specimen(spcid) for spcid in self._SPECIMEN_IDS
            }

        if AllenBrainAtlasQuery.factors is None:
            self._retrieve_factors()

        # get expression levels and z_scores for the gene
        if len(probe_ids) > 0:
            for donor_id in self._DONOR_IDS:
                for item in self._retrieve_microarray(donor_id, probe_ids):
                    yield item

    @staticmethod
    def _retrieve_probe_ids(genes: list):
        assert isinstance(genes, list)
        if len(genes) == 1:
            logger.debug(f"Retrieving probe ids for gene {genes[0]['symbol']}")
        else:
            logger.debug(f"Retrieving probe ids for genes {', '.join(g['symbol'] for g in genes)}")
        start_row = 0
        num_rows = 50
        probe_ids = []
        while True:
            url = AllenBrainAtlasQuery._QUERY["multiple_gene_probe"].format(
                start_row=start_row, num_rows=num_rows, genes=','.join([f"'{g['symbol']}'" for g in genes])
            )
            response = HttpRequest(url).get()
            if "site unavailable" in response.decode().lower():
                # When the Allen site is not available, they still send a status code 200.
                raise RuntimeError(
                    "Allen institute site unavailable - please try again later."
                )
            root = ElementTree.fromstring(response)
            num_probes = int(root.attrib["num_rows"])
            total_probes = int(root.attrib["total_rows"])
            assert len(root) == 1
            probe_ids.extend([int(root[0][i][0].text) for i in range(num_probes)])
            if (start_row + num_rows) >= total_probes:
                break
            # retrieve another page
            start_row += num_rows
        return probe_ids

    @staticmethod
    def _retrieve_factors():
        start_row = 0
        num_rows = 50
        if AllenBrainAtlasQuery.factors is None or len(AllenBrainAtlasQuery.factors) == 0:
            AllenBrainAtlasQuery.factors = {}
            while True:
                factors_url = AllenBrainAtlasQuery._QUERY["factors"].format(start_row=start_row, num_rows=num_rows)
                response = HttpRequest(factors_url).get()
                AllenBrainAtlasQuery.factors.update({
                    item["id"]: {
                        "race": item["race_only"],
                        "gender": item["sex"],
                        "age": int(item["age"]["days"] / 365),
                    }
                    for item in response["msg"]
                })
                total_factors = int(response["total_rows"])
                if (start_row + num_rows) >= total_factors:
                    break
                # retrieve another page
                start_row += num_rows

    @staticmethod
    def _retrieve_specimen(specimen_id: str):
        """
        Retrieves information about a human specimen.
        """
        url = AllenBrainAtlasQuery._QUERY["specimen"].format(specimen_id=specimen_id)
        response = HttpRequest(url).get()
        if not response["success"]:
            raise InvalidAllenAPIResponseException(
                "Invalid response when retrieving specimen information: {}".format(url)
            )
        # we ask for 1 specimen, so list should have length 1
        assert len(response["msg"]) == 1
        specimen = response["msg"][0]
        T = specimen["alignment3d"]
        specimen["donor2icbm"] = np.array(
            [
                [T["tvr_00"], T["tvr_01"], T["tvr_02"], T["tvr_09"]],
                [T["tvr_03"], T["tvr_04"], T["tvr_05"], T["tvr_10"]],
                [T["tvr_06"], T["tvr_07"], T["tvr_08"], T["tvr_11"]],
            ]
        )
        return specimen

    @classmethod
    def _retrieve_microarray(cls, donor_id: str, probe_ids: str):
        """
        Retrieve microarray data for several probes of a given donor, and
        compute the MRI position of the corresponding tissue block in the ICBM
        152 space to generate a SpatialFeature object for each sample.
        """

        if len(probe_ids) == 0:
            return

        # query the microarray data for this donor
        url = AllenBrainAtlasQuery._QUERY["microarray"].format(
            probe_ids=",".join([str(id) for id in probe_ids]), donor_id=donor_id
        )
        try:
            response = HttpRequest(url, json.loads).get()
        except json.JSONDecodeError as e:
            raise RuntimeError(f"Allen institute site produced an empty response - please try again later.\n{e}")
        if not response["success"]:
            raise InvalidAllenAPIResponseException(
                "Invalid response when retrieving microarray data: {}".format(url)
            )

        probes, samples = [response["msg"][n] for n in ["probes", "samples"]]

        for i, sample in enumerate(samples):

            # coordinate conversion to ICBM152 standard space
            donor_id = sample["donor"]["id"]
            donor_name = sample["donor"]["name"]
            icbm_coord = (np.matmul(
                AllenBrainAtlasQuery._specimen[donor_name]["donor2icbm"],
                sample["sample"]["mri"] + [1],
            )).round(2)

            for probe in probes:
                yield {
                    "gene": probe['gene-symbol'],
                    "expression_level": float(probe["expression_level"][i]),
                    "z_score": float(probe["z-score"][i]),
                    "sample_index": i,
                    "probe_id": probe["id"],
                    "donor_id": donor_id,
                    "donor_name": donor_name,
                    "race": cls.factors[donor_id]["race"],
                    "gender": cls.factors[donor_id]["gender"],
                    "age": cls.factors[donor_id]["age"],
                    "mni_xyz": icbm_coord[:3],
                }



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Multimodal data features in 2D section."""

from typing import TYPE_CHECKING

from . import image

if TYPE_CHECKING:
    from ...locations import AxisAlignedPatch, Contour
    from ...features.anchor import AnatomicalAnchor


class CellbodyStainedSection(
    image.Image,
    configuration_folder="features/images/sections/cellbody",
    category="cellular",
):
    def __init__(self, **kwargs):
        image.Image.__init__(self, **kwargs, modality="cell body staining")

    def get_boundingbox(self, clip=False, background=0, **fetch_kwargs):
        return super().get_boundingbox(clip, background, **fetch_kwargs)


class BigBrain1MicronPatch(image.Image, category="cellular"):

    _DESCRIPTION = """Sample approximately orthogonal cortical image patches
    from BigBrain 1 micron sections, guided by an image volume
    in a supported reference space providing. The image
    volume is used as a weighted mask to extract patches
    along the cortical midsurface with nonzero weights in the
    input image.
    An optional lower_threshold can be used to narrow down
    the search. The weight is stored with the resulting features."""

    def __init__(
        self,
        patch: "AxisAlignedPatch",
        profile: "Contour",
        section: CellbodyStainedSection,
        vertex: int,
        relevance: float,
        anchor: "AnatomicalAnchor",
    ):
        self._patch = patch
        self._profile = profile
        self._section = section
        self.vertex = vertex
        self.relevance = relevance
        image.Image.__init__(
            self,
            name=f"Cortical patch in {section.name}",
            modality=section.modality,
            space_spec=section._space_spec,
            providers=list(section._providers.values()),
            region=None,
            datasets=section.datasets,
            bbox=patch.boundingbox,
            id=None,
        )
        self._anchor_cached = anchor
        self._description_cached = self._DESCRIPTION

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}(space_spec={self._space_spec}, "
            f"name='{self.name}', "
            f"section='{self._section.get_boundingbox().minpoint.bigbrain_section()}', "
            f"vertex='{self.vertex}', providers={self._providers})>"
        )

    @property
    def section(self) -> CellbodyStainedSection:
        return self._section

    def get_boundingbox(self, **fetch_kwargs):
        """Enforce that the bounding box spans the full section thickness."""
        bbox_section = self._section.get_boundingbox(**fetch_kwargs)
        bbox = self._patch.boundingbox
        bbox.minpoint[1] = bbox_section.minpoint[1]
        bbox.maxpoint[1] = bbox_section.maxpoint[1]
        return bbox

    @property
    def profile(self) -> "Contour":
        return self._profile

    @property
    def bigbrain_section(self):
        return self.get_boundingbox().minpoint.bigbrain_section()

    def fetch(self, flip: bool = False, resolution_mm: float = -1, **kwargs):
        assert len(kwargs) == 0
        p = self._patch.flip() if flip else self._patch
        return p.extract_volume(self._section, resolution_mm=resolution_mm).fetch()
